###########################################################################################################################################################################
# CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR CNES_SR #
###########################################################################################################################################################################

import psycopg2


"""
Cria o schema do sub banco de dados do CNES_SR (Serviços Especializados) para atender ao formato do SGBD postgreSQL.
Na verdade o schema abaixo deve atender ao formato dos SGBD Oracle e SQL Server com no máximo pequenas modificações.

"""

# Função de criação das tabelas do sub banco de dados cnes_sr
def create_tables(connection_data, child_db):
    conn = psycopg2.connect(dbname=connection_data[0],
                            host=connection_data[1],
                            port=connection_data[2],
                            user=connection_data[3],
                            password=connection_data[4])

    cursor = conn.cursor()

    cursor.execute(f'''/*Tabela Serviços Especializados (tabela principal do sub banco de dados cnes_sr)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.srbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_SR"                   VARCHAR(2),
                                                                  "ANO_SR"                  INTEGER,
                                                                  "MES_SR"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "SERVESP_ID"              VARCHAR(3),
                                                                  "CLASSSR_ID"              VARCHAR(6),
                                                                  "SRVUNICO_ID"             VARCHAR(3),
                                                                  "TPGESTAO_ID"             VARCHAR(2),
                                                                  "PFPJ_ID"                 VARCHAR(2),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "NIVDEP_ID"               VARCHAR(2),
                                                                  "ESFERAA_ID"              VARCHAR(2),
                                                                  "ATIVIDAD_ID"             VARCHAR(2),
                                                                  "RETENCAO_ID"             VARCHAR(2),
                                                                  "NATUREZA_ID"             VARCHAR(2),
                                                                  "CLIENTEL_ID"             VARCHAR(2),
                                                                  "TPUNID_ID"               VARCHAR(2),
                                                                  "TURNOAT_ID"              VARCHAR(2),
                                                                  "NIVHIER_ID"              VARCHAR(2),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "CNPJ_MAN"                VARCHAR(14),
                                                                  "CARACTER_ID"             VARCHAR(2),
                                                                  "AMB_NSUS"                NUMERIC,
                                                                  "AMB_SUS"                 NUMERIC,
                                                                  "HOSP_NSUS"               NUMERIC,
                                                                  "HOSP_SUS"                NUMERIC,
                                                                  "CONTSRVU"                NUMERIC,
                                                                  "CNESTERC"                VARCHAR(7),
                                                                  "NATJUR_ID"               VARCHAR(4));

                       /*Tabela dos estabelecimentos de saúde*/
                       CREATE TABLE IF NOT EXISTS {child_db}.cnes("ID"                      VARCHAR(7),
                                                                  "DESCESTAB"               VARCHAR(66),
                                                                  "RSOC_MAN"                VARCHAR(66),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "EXCLUIDO"                NUMERIC,
                                                                  "DATAINCL"                DATE,
                                                                  "DATAEXCL"                DATE);

                       /*Tabela dos municípios de localização de estabelecimentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.codufmun("ID"                  VARCHAR(6),
                                                                      "MUNNOME"	            VARCHAR(66),
                                                                      "MUNNOMEX"            VARCHAR(66),
                                                                      "MUNCODDV"            VARCHAR(7),
                                                                      "OBSERV"              VARCHAR(66),
                                                                      "SITUACAO"            VARCHAR(66),
                                                                      "MUNSINP"	            VARCHAR(66),
                                                                      "MUNSIAFI"            VARCHAR(66),
                                                                      "UFCOD_ID"            VARCHAR(2),
                                                                      "AMAZONIA"            VARCHAR(66),
                                                                      "FRONTEIRA"           VARCHAR(66),
                                                                      "CAPITAL"             VARCHAR(66),
                                                                      "RSAUDE_ID"           VARCHAR(5),
                                                                      "LATITUDE"            FLOAT,
                                                                      "LONGITUDE"           FLOAT,
                                                                      "ALTITUDE"            FLOAT,
                                                                      "AREA"                FLOAT,
                                                                      "ANOINST"             VARCHAR(66),
                                                                      "ANOEXT"              VARCHAR(66),
                                                                      "SUCESSOR"            VARCHAR(66));

                       /*Tabela dos serviços especializados*/
                       CREATE TABLE IF NOT EXISTS {child_db}.servesp("ID"                   VARCHAR(3),
                                                                     "DESCRICAO"            VARCHAR(100));

                       /*Tabela de classificações dos serviços especializados*/
                       CREATE TABLE IF NOT EXISTS {child_db}.classsr("ID"                   VARCHAR(6),
                                                                     "DESCRICAO"            VARCHAR(100));

                       /*Na verdade é a mesma coisa da tabela servesp*/
                       CREATE TABLE IF NOT EXISTS {child_db}.srvunico("ID"                  VARCHAR(3),
                                                                      "DESCRICAO"           VARCHAR(100));

                       /*Tabela dos tipos de gestão*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpgestao("ID"                  VARCHAR(2),
                                                                      "GESTAO"              VARCHAR(66));

                       /*Tabela se o estabelecimento é pessoa física ou pessoa jurídica*/
                       CREATE TABLE IF NOT EXISTS {child_db}.pfpj("ID"                      VARCHAR(2),
                                                                  "PESSOA"                  VARCHAR(66));

                       /*Tabela do grau de independência do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.nivdep("ID"                    VARCHAR(2),
                                                                    "TIPO"                  VARCHAR(66));

                       /*Tabela dos tipos de administração*/
                       CREATE TABLE IF NOT EXISTS {child_db}.esferaa("ID"                   VARCHAR(2),
                                                                     "ADMINISTRACAO"        VARCHAR(66));

                       /*Tabela dos tipos de atividade de ensino, se houver*/
                       CREATE TABLE IF NOT EXISTS {child_db}.atividad("ID"                  VARCHAR(2),
                                                                      "ATIVIDADE"           VARCHAR(66));

                       /*Tabela dos tipos de retenção de tributos do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.retencao("ID"                  VARCHAR(2),
                                                                      "RETENCAO"            VARCHAR(66));

                       /*Tabela da natureza do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.natureza("ID"                  VARCHAR(2),
                                                                      "NATUREZA"            VARCHAR(66));

                       /*Tabela dos tipos de fluxo de clientela*/
                       CREATE TABLE IF NOT EXISTS {child_db}.clientel("ID"                  VARCHAR(2),
                                                                      "CLIENTELA"           VARCHAR(66));

                       /*Tabela dos tipos de estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpunid("ID"                    VARCHAR(2),
                                                                    "TIPO"                  VARCHAR(66));

                       /*Tabela dos turnos de funcionamento do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.turnoat("ID"                   VARCHAR(2),
                                                                     "TURNO"                VARCHAR(66));

                       /*Tabela dos níveis de atendimento do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.nivhier("ID"                   VARCHAR(2),
                                                                     "NIVEL"                VARCHAR(66));

                       /*Tabela de caracterizações do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.caracter("ID"                  VARCHAR(2),
                                                                      "CARACTERIZACAO"      VARCHAR(66));

                       /*Tabela das naturezas jurídicas de estabelecimentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.natjur("ID"                    VARCHAR(4),
                                                                    "NATUREZA"              VARCHAR(100));

                       /*Tabela dos Estados da RFB*/
                       CREATE TABLE IF NOT EXISTS {child_db}.ufcod("ID"                     VARCHAR(2),
                                                                   "ESTADO"                 VARCHAR(66),
                                                                   "SIGLA_UF"               VARCHAR(66));

                       /*Tabela de regiões de saúde IBGE*/
                       CREATE TABLE IF NOT EXISTS {child_db}.rsaude("ID"                    VARCHAR(5),
                                                                    "REGIAO"                VARCHAR(66));

                       /*Tabela de Arquivos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.arquivos("NOME"                VARCHAR(15),
                                                                      "DIRETORIO"           VARCHAR(66),
                                                                      "DATA_INSERCAO_FTP"   DATE,
                                                                      "DATA_HORA_CARGA"     TIMESTAMP,
                                                                      "QTD_REGISTROS"       INTEGER);

                ''')
    conn.commit()
    # Encerra o cursor
    cursor.close()
    # Encerra a conexão
    conn.close()



###########################################################################################################################
# SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA #
###########################################################################################################################

import os
from datetime import datetime

import numpy as np
import pandas as pd

from transform.extract.download_SIA import download_SIAXXaamm, download_table_dbf, download_table_cnv

"""
Módulo de limpeza/tratamento de dados do SIA.

"""

# Função para converter um "value" num certo "type" de objeto ou caso não seja possível utiliza o valor "default"
def tryconvert(value, default, type):
    try:
        return type(value)
    except (ValueError, TypeError):
        return default


# Classe de dados principais do SIA
class DataSiaMain:

    # Construtor
    def __init__(self, base, state, year, month):
        self.base = base
        self.state = state
        self.year = year
        self.month = month


    # Método para ler como um objeto pandas DataFrame um arquivo principal de dados do SIA e adequar e formatar suas...
    # colunas e valores
    def get_SIAXXaamm_treated(self):
        # Lê o arquivo "dbc" ou "parquet", se já tiver sido baixado, como um objeto pandas DataFrame
        dataframe = download_SIAXXaamm(self.base, self.state, self.year, self.month)
        print(f'O número de linhas do arquivo {self.base}{self.state}{self.year}{self.month} é {dataframe.shape[0]}.')

        ###################################################################################################################
        # SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA #
        ###################################################################################################################
        if self.base == 'PA':
            # Colunas definidas como necessárias no objeto pandas DataFrame que incrementará a tabela pabr da base de dados
            lista_columns = np.array(['PA_CODUNI', 'PA_GESTAO', 'PA_CONDIC', 'PA_UFMUN', 'PA_REGCT', 'PA_INCOUT',
                                      'PA_INCURG', 'PA_TPUPS', 'PA_TIPPRE', 'PA_MN_IND', 'PA_CNPJCPF', 'PA_CNPJMNT',
                                      'PA_CNPJ_CC', 'PA_MVM', 'PA_CMP', 'PA_PROC_ID', 'PA_TPFIN', 'PA_NIVCPL',
                                      'PA_DOCORIG', 'PA_AUTORIZ', 'PA_CNSMED', 'PA_CBOCOD', 'PA_MOTSAI', 'PA_OBITO',
                                      'PA_ENCERR', 'PA_PERMAN', 'PA_ALTA', 'PA_TRANSF', 'PA_CIDPRI', 'PA_CIDSEC',
                                      'PA_CIDCAS', 'PA_CATEND', 'PA_IDADE', 'IDADEMIN', 'IDADEMAX', 'PA_FLIDADE',
                                      'PA_SEXO', 'PA_RACACOR', 'PA_MUNPCN', 'PA_QTDPRO', 'PA_QTDAPR', 'PA_VALPRO',
                                      'PA_VALAPR', 'PA_UFDIF', 'PA_MNDIF', 'PA_DIF_VAL', 'NU_VPA_TOT', 'NU_PA_TOT',
                                      'PA_INDICA', 'PA_CODOCO', 'PA_FLQT', 'PA_FLER', 'PA_ETNIA', 'PA_VL_CF',
                                      'PA_VL_CL', 'PA_VL_INC', 'PA_SRC_C', 'PA_INE', 'PA_NAT_JUR'])

            # Criação de um objeto pandas DataFrame vazio com as colunas especificadas acima
            df = pd.DataFrame(columns=lista_columns)

            # Colocação dos dados da variável "dataframe" na variável "df" nas colunas de mesmo nome preenchendo...
            # automaticamente com o float NaN as colunas da variável "df" não presentes na variável dataframe
            for col in df.columns.values:
                for coluna in dataframe.columns.values:
                    if coluna == col:
                        df[col] = dataframe[coluna].tolist()
                        break

            # Coloca na variável "dif_set" o objeto array dos nomes das colunas da variável "df" que não estão...
            # presentes na variável "dataframe"
            dif_set = np.setdiff1d(df.columns.values, dataframe.columns.values)

            # Substitui o float NaN pela string vazia as colunas da variável "df" não presentes na variável "dataframe"
            for col in dif_set:
                df[col].replace(np.nan, '', inplace=True)

            # Inserção de coluna de grupo, subgrupo e forma do procedimento realizado
            df['GRUPO'] = df['PA_PROC_ID'].apply(lambda x: x[:2])
            df['SUBGRUPO'] = df['PA_PROC_ID'].apply(lambda x: x[:4])
            df['FORMA'] = df['PA_PROC_ID'].apply(lambda x: x[:6])


            # Simplifica/corrige a apresentação dos dados das colunas especificadas
            df['PA_CODUNI'] = df['PA_CODUNI'].apply(lambda x: x.zfill(7))
            df['PA_CODUNI'] = df['PA_CODUNI'].apply(str.strip)
            df['PA_CODUNI'] = df['PA_CODUNI'].apply(lambda x: x if len(x) == 7 else '')

            for col in np.array(['PA_REGCT', 'PA_INCOUT', 'PA_INCURG']):
                df[col] = df[col].apply(lambda x: x.zfill(4))
                df[col] = df[col].apply(str.strip)

            for col in np.array(['PA_TPUPS', 'PA_TPFIN', 'PA_CATEND', 'PA_RACACOR']):
                for i in np.array(['00', '01', '02', '03', '04', '05', '06', '07', '08', '09']):
                    df[col].replace(i, str(int(i)), inplace=True)

            df['PA_AUTORIZ'].replace(to_replace=',', value= '', regex=True, inplace=True)

            for col in np.array(['PA_CBOCOD', 'PA_SRC_C']):
                df[col] = df[col].apply(lambda x: x.zfill(6))
                df[col] = df[col].apply(str.strip)
                df[col] = df[col].apply(lambda x: x if len(x) == 6 else '')

            for col in np.array(['PA_CIDPRI', 'PA_CIDSEC', 'PA_CIDCAS']):
                df[col] = df[col].apply(lambda x: x[:4] if len(x) > 4 else x)

            df['PA_CODOCO'] = df['PA_CODOCO'] + df['PA_FLQT'] # Fusão em uma coluna
            df = df.drop(['PA_FLQT'], axis=1) # Drop a coluna PA_FLQT

            df['PA_ETNIA'] = df['PA_ETNIA'].apply(lambda x: x.zfill(4))
            df['PA_ETNIA'] = df['PA_ETNIA'].apply(str.strip)
            df['PA_ETNIA'] = df['PA_ETNIA'].apply(lambda x: x if len(x) == 4 else '')

            # Atualiza/corrige os labels das colunas especificadas
            for col in np.array(['PA_GESTAO', 'PA_UFMUN', 'PA_MUNPCN']):
                df[col] = df[col].apply(lambda x: x if len(x) == 6 else '')
                df[col].replace([str(i) for i in range(334501, 334531)], '330455', inplace=True)
                df[col].replace([str(i) for i in range(358001, 358059)], '355030', inplace=True)
                df[col].replace(['000000', '150475', '421265', '422000', '431454',
                                 '500627', '510445', '999999'], '', inplace=True)
                df[col].replace(['530000', '530020', '530030', '530040', '530050',
                                 '530060', '530070', '530080', '530090', '530100',
                                 '530110', '530120', '530130', '530135', '530140',
                                 '530150', '530160','530170', '530180'] + \
                                 [str(i) for i in range(539900, 540000)], '530010', inplace=True)

            df['PA_REGCT'].replace(['7114'], '', inplace=True)

            for col in np.array(['PA_INCOUT', 'PA_INCURG']):
                df[col].replace(['0001', '9999'], '1', inplace=True) # "1" de "Sim"
                df[col].replace('0000', '0', inplace=True)           # "0" de "Não"

            df['PA_TPUPS'].replace(['82', '84', '85', '99'], '', inplace=True)

            df['PA_TIPPRE'].replace('00', '', inplace=True)

            for col in np.array(['PA_MVM', 'PA_CMP']):
                df[col] = df[col].apply(lambda x: x if len(x) == 6 else '')
                df[col] = df[col].apply(lambda x: x if 1 <= tryconvert(x[4:6], 0, int) <= 12 else '')

            for col in np.array(['PA_CBOCOD', 'PA_SRC_C']):
                df[col].replace(['000000', '515140', '5151F1'], '', inplace=True)

            df['PA_MOTSAI'].replace('00', '0', inplace=True)

            for col in np.array(['PA_CIDPRI', 'PA_CIDSEC', 'PA_CIDCAS']):
                df[col] = df[col].apply(lambda x: x.upper())
                df[col].replace(['0000', '9999', 'C200', 'C550', 'C560', 'C610', 'C640',
                                 'C800', 'C970', 'G455', 'J460', 'J462', 'J464', 'J465',
                                 'J466', 'J467', 'J470', 'J481', 'J484', 'R007', 'R495'], '', inplace=True)
                df[col].replace('B501', 'B508', inplace=True)
                df[col].replace('B656', 'B653', inplace=True)
                df[col].replace('C141', 'C140', inplace=True)
                df[col].replace('M723', 'M724', inplace=True)
                df[col].replace('M725', 'M728', inplace=True)
                df[col].replace('N975', 'N978', inplace=True)
                df[col].replace('Q314', 'P288', inplace=True)
                df[col].replace('Q350', 'Q351', inplace=True)
                df[col].replace('Q352', 'Q353', inplace=True)
                df[col].replace('Q354', 'Q355', inplace=True)
                df[col].replace(['Q356', 'Q358'], 'Q359', inplace=True)
                df[col].replace('R500', 'R508', inplace=True)
                df[col].replace('R501', 'R500', inplace=True)
                df[col].replace(['X590', 'X591', 'X592', 'X593', 'X594', 'X595',
                                 'X596', 'X597', 'X598'], 'X599', inplace=True)
                df[col].replace('Y34', 'Y349', inplace=True)
                df[col].replace('Y447', 'Y448', inplace=True)
                df[col].replace('U04', '', inplace=True)

            df['PA_CATEND'].replace(['0', '7', '10', '12', '20', '53', '54',
                                     '57', '0-', '0E', '0U'], '', inplace=True)

            df['PA_RACACOR'].replace(['0', '6', '9', '1M', '1G', '1C',
                                      'DE', 'D', '87', '99'], '', inplace=True)

            df['PA_ETNIA'].replace(['0000', '9999'], '', inplace=True)
            df['PA_ETNIA'].replace(regex='^X.{3}', value='', inplace=True)

            df['PA_INE'].replace(['0001610902', '0001619055', '0001630520',
                                  '0001627112', '0001636383'], '', inplace=True)

            df['PA_NAT_JUR'].replace(['0000', '1244'], '', inplace=True)

            # Substitui uma string vazia pela string "NA" nas colunas de foreign keys
            for col in np.array(['PA_CODUNI', 'PA_GESTAO', 'PA_CONDIC', 'PA_UFMUN', 'PA_REGCT',
                                 'PA_TPUPS', 'PA_TIPPRE',  'PA_PROC_ID', 'PA_TPFIN', 'PA_NIVCPL',
                                 'PA_DOCORIG', 'PA_CBOCOD', 'PA_MOTSAI', 'PA_CIDPRI', 'PA_CIDSEC',
                                 'PA_CIDCAS', 'PA_CATEND', 'PA_FLIDADE', 'PA_SEXO', 'PA_RACACOR',
                                 'PA_MUNPCN', 'PA_INDICA', 'PA_CODOCO', 'PA_ETNIA', 'PA_SRC_C',
                                 'PA_INE', 'PA_NAT_JUR', 'GRUPO', 'SUBGRUPO', 'FORMA']):
                df[col].replace('', 'NA', inplace=True)

            # Substitui uma string vazia por None nas colunas de atributos especificadas
            for col in np.array(['PA_MN_IND', 'PA_CNPJCPF', 'PA_CNPJMNT',
                                 'PA_CNPJ_CC', 'PA_AUTORIZ', 'PA_CNSMED']):
                df[col].replace('', None, inplace=True)

            # Converte do tipo string para datetime as colunas especificadas substituindo as datas faltantes...
            # ("NaT") pela data futura "2099-01-01" para permitir a inserção das referidas colunas no SGBD postgreSQL
            for col in np.array(['PA_MVM', 'PA_CMP']):
                df[col] = df[col].apply(lambda x: datetime.strptime(x, '%Y%m').date() \
                                        if x != '' else datetime(2099, 1, 1).date())

            # Verifica se as datas das colunas especificadas são absurdas e em caso afirmativo as substitui...
            # pela data futura "2099-01-01"
            for col in np.array(['PA_MVM', 'PA_CMP']):
                df[col] = df[col].apply(lambda x: x if datetime(1850, 12, 31).date() < x < \
                                        datetime(2020, 12, 31).date() else datetime(2099, 1, 1).date())

            # Converte do tipo object para int ou para None as colunas de atributos de valores binários...
            # (0 ou 1) ou quando simplesmente se deseja inteiros
            for col in np.array(['PA_INCOUT', 'PA_INCURG', 'PA_OBITO', 'PA_ENCERR', 'PA_PERMAN', 'PA_ALTA', 'PA_TRANSF',
                                 'PA_IDADE', 'IDADEMIN', 'IDADEMAX', 'PA_QTDPRO', 'PA_QTDAPR', 'PA_UFDIF', 'PA_FLER']):
                df[col] = df[col].apply(lambda x: tryconvert(x, None, int))

            # Converte do tipo object para float com duas casas decimais as colunas de atributos de valores...
            # representativos de quantidades ou para o valor None caso a coluna esteja com a string vazia
            for col in np.array(['PA_VALPRO', 'PA_VALAPR', 'PA_DIF_VAL', 'NU_VPA_TOT',
                                 'NU_PA_TOT', 'PA_VL_CF','PA_VL_CL', 'PA_VL_INC']):
                df[col] = df[col].apply(lambda x: round(float(x), 2) if x != '' else None)

            # Renomeia colunas que são foreign keys
            df.rename(index=str, columns={'PA_CODUNI': 'PACODUNI_ID', 'PA_GESTAO': 'PAGESTAO_ID',
                                          'PA_CONDIC': 'PACONDIC_ID', 'PA_UFMUN': 'PAUFMUN_ID',
                                          'PA_REGCT': 'PAREGCT_ID', 'PA_TPUPS': 'PATPUPS_ID',
                                          'PA_TIPPRE': 'PATIPPRE_ID', 'PA_PROC_ID': 'PAPROC_ID',
                                          'PA_TPFIN': 'PATPFIN_ID', 'PA_NIVCPL': 'PANIVCPL_ID',
                                          'PA_DOCORIG': 'PADOCORIG_ID', 'PA_CBOCOD': 'PACBOCOD_ID',
                                          'PA_MOTSAI': 'PAMOTSAI_ID', 'PA_CIDPRI': 'PACIDPRI_ID',
                                          'PA_CIDSEC': 'PACIDSEC_ID', 'PA_CIDCAS': 'PACIDCAS_ID',
                                          'PA_CATEND': 'PACATEND_ID', 'PA_FLIDADE': 'PAFLIDADE_ID',
                                          'PA_SEXO': 'PASEXO_ID', 'PA_RACACOR': 'PARACACOR_ID',
                                          'PA_MUNPCN': 'PAMUNPCN_ID', 'PA_INDICA': 'PAINDICA_ID',
                                          'PA_CODOCO': 'PACODOCO_ID', 'PA_ETNIA': 'PAETNIA_ID',
                                          'PA_SRC_C': 'PASRCC_ID', 'PA_INE': 'PAINE_ID',
                                          'PA_NAT_JUR': 'PANATJUR_ID', 'GRUPO': 'GRUPO_ID',
                                          'SUBGRUPO': 'SUBGRUPO_ID', 'FORMA': 'FORMA_ID'}, inplace=True)

            print(f'Tratou o arquivo PA{self.state}{self.year}{self.month} (shape final: {df.shape[0]} x {df.shape[1]}).')

            return df


# Classe de dados auxiliares do SIA
class DataSiaAuxiliary:

    # Construtor
    def __init__(self, path):
        self.path = path


    #######################################################################################################################
    # SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA #
    #######################################################################################################################


    ###################################################################################################################
    # SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA SIA_PA #
    ###################################################################################################################
    # Função para adequar e formatar as colunas e valores das 27 Tabelas CADGERXX (arquivos CADGERXX.dbf),...
    # sendo uma para cada estado do Brasil
    def get_CADGERBR_treated(self):
        # Conversão das 27 Tabelas CADGERXX para um objeto pandas DataFrame
        frames = []
        lista_estados = ['AC', 'AL', 'AM', 'AP', 'BA', 'CE', 'DF', 'ES', 'GO', 'MA', 'MG', 'MS', 'MT',
                         'PA', 'PB', 'PE', 'PI', 'PR', 'RJ', 'RN', 'RO', 'RR', 'RS', 'SC', 'SE', 'SP', 'TO']
        for state in lista_estados:
            file_name = 'CADGER' + state
            dfi = download_table_dbf(file_name)
            # Renomeia as colunas especificadas
            dfi.rename(index=str, columns={'CNES': 'ID', 'FANTASIA': 'DESCESTAB'}, inplace=True)
            # Remove colunas indesejáveis do objeto pandas DataFrame
            dfi = dfi.drop(['RAZ_SOCI', 'LOGRADOU', 'NUM_END', 'COMPLEME', 'BAIRRO',
                            'COD_CEP', 'TELEFONE',  'FAX', 'EMAIL', 'REGSAUDE', 'MICR_REG',
                            'DISTRSAN', 'DISTRADM', 'CODUFMUN', 'NATUREZA'], axis=1)
            # Reordena as colunas
            dfi = dfi[['ID', 'DESCESTAB', 'RSOC_MAN', 'CPF_CNPJ', 'EXCLUIDO', 'DATAINCL', 'DATAEXCL']]
            # Elimina linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
            dfi.drop_duplicates(subset='ID', keep='first', inplace=True)
            # Ordena as linhas por ordem crescente dos valores da coluna ID
            dfi.sort_values(by=['ID'], inplace=True)
            # Reset o index devido ao sorting prévio e à eventual eliminação de duplicates
            dfi.reset_index(drop=True, inplace=True)
            frames.append(dfi)
        df = pd.concat(frames, ignore_index=True)
        # Elimina linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        df.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena as linhas por ordem crescente dos valores da coluna ID
        df.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à eventual eliminação de duplicates
        df.reset_index(drop=True, inplace=True)
        # Converte do tipo object para int ou para None as colunas de atributos de valores binários (0 ou 1)
        df['EXCLUIDO'] = df['EXCLUIDO'].apply(lambda x: tryconvert(x, None, int))
        # Substitui o objeto datetime.date "9999-12-31" das duas colunas de "datas" pelo objeto...
        # datetime.date "2099-01-01"
        for col in np.array(['DATAINCL', 'DATAEXCL']):
            df[col].replace(datetime(9999, 12, 31).date(), datetime(2099, 1, 1).date(), inplace=True)
        # Verifica se as datas das colunas especificadas são absurdas e em caso afirmativo as substitui pela...
        # data futura "2099-01-01"
        for col in np.array(['DATAINCL', 'DATAEXCL']):
            df[col] = df[col].apply(lambda x: x if datetime(1850, 12, 31).date() < x < \
                                    datetime(2020, 12, 31).date() else datetime(2099, 1, 1).date())

        # Upload do arquivo "xlsx" que contém os CNES presentes nos arquivos PAXXaamm (dos anos de...
        # 2008 a 2019) e não presentes nas 27 Tabelas CADGERXX. Ou seja,...
        # isso parece ser uma falha dos dados do Datasus
        dataframe = pd.read_excel(self.path + 'PA_CODUNI_OUT_CADGER_XX_ANOS_2008_2019' + '.xlsx')
        # Converte a coluna "ID" do objeto "dataframe" de "int" para "string"
        dataframe['ID'] = dataframe['ID'].astype('str')
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "dataframe" até...
        # formar uma "string" de tamanho = 7
        dataframe['ID'] = dataframe['ID'].apply(lambda x: x.zfill(7))
        # Adiciona as colunas DESCESTAB, RSOC_MAN, CPF_CNPJ, EXCLUIDO, DATAINCL e DATAEXCL e respectivos...
        # valores ao objeto "dataframe"
        dataframe['DESCESTAB'] = 'NAO PROVIDO EM 27 ARQUIVOS DBF DE CNES'
        dataframe['RSOC_MAN'] = '?'
        dataframe['CPF_CNPJ'] = '?'
        dataframe['EXCLUIDO'] = None
        dataframe['DATAINCL'] = datetime(2099, 1, 1).date()
        dataframe['DATAEXCL'] = datetime(2099, 1, 1).date()
        # Concatenação do objeto "dataframe" ao objeto "df"
        frames = []
        frames.append(df)
        frames.append(dataframe)
        dfinal = pd.concat(frames, ignore_index=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        dfinal.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena eventualmente as linhas por ordem crescente dos valores da coluna ID
        dfinal.sort_values(by=['ID'], inplace=True)
        # Reset eventualmente o index devido ao sorting prévio e à eventual eliminação de duplicates
        dfinal.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        dfinal.loc[dfinal.shape[0]] = ['NA', 'NOT AVAILABLE', '?', '?', None, datetime(2099, 1, 1), datetime(2099, 1, 1)]
        return dfinal


    # Método para adequar e formatar as colunas e valores da Tabela CADMUN (arquivo CADMUN.dbf)
    def get_CADMUN_treated(self):
        # Conversão da Tabela CADMUN para um objeto pandas DataFrame
        file_name = 'CADMUN'
        df1 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df1.rename(index=str, columns={'MUNCOD': 'ID', 'UFCOD': 'UFCOD_ID'}, inplace=True)
        # Drop a linha inteira em que a coluna "ID" tem o valor especificado por não representar nenhum município
        df1 = df1.drop(df1[df1['ID']=='000000'].index)
        # Remove colunas indesejáveis do objeto pandas DataFrame
        df1 = df1.drop(['MUNSINON', 'MUNSINONDV', 'MESOCOD', 'MICROCOD', 'MSAUDCOD',
                        'RSAUDCOD', 'CSAUDCOD', 'RMETRCOD', 'AGLCOD'], axis=1)
        # Substitui uma string vazia pela string "?" nas colunas especificadas
        for col in ['SITUACAO', 'MUNSINP', 'MUNSIAFI', 'MUNNOME', 'MUNNOMEX', 'OBSERV',
                    'AMAZONIA', 'FRONTEIRA', 'CAPITAL', 'ANOINST', 'ANOEXT', 'SUCESSOR']:
            df1[col].replace('', '?', inplace=True)
        # Substitui uma string vazia pela string "NA" nas colunas especificadas
        df1['UFCOD_ID'].replace('', 'NA', inplace=True)
        # Substitui uma string vazia pelo float "NaN" nas colunas especificadas
        for col in ['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']:
            df1[col].replace('', np.nan, inplace=True)
        # Converte do tipo object para float as colunas especificadas
        df1[['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']] = \
        df1[['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']].astype('float')
        # Coloca todas as string das colunas especificadas como UPPER CASE
        df1['MUNNOME'] = df1['MUNNOME'].apply(lambda x: x.upper())
        df1['MUNNOMEX'] = df1['MUNNOMEX'].apply(lambda x: x.upper())
        # Insere uma linha referente ao Município de Nazária/PI não constante originalmente do arquivo
        df1.loc[df1.shape[0]] = ['220672', '2206720', 'ATIVO', '?', '?', 'NAZÁRIA', 'NAZARIA', '?',
                                 'N', 'N', 'N', '22', '?', '?', '?', np.nan, np.nan, np.nan, 363.589]
        # Ordena as linhas de "df1" por ordem crescente dos valores da coluna ID
        df1.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à exclusão e inclusão das linhas referidas acima
        df1.reset_index(drop=True, inplace=True)
        # Conversão da Tabela rl_municip_regsaud para um objeto pandas DataFrame
        file_name = 'rl_municip_regsaud'
        df2 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df2.rename(index=str, columns={'CO_MUNICIP': 'ID', 'CO_REGSAUD': 'RSAUDE_ID'}, inplace=True)
        # Faz o merge de "df1" e "df2" pela coluna ID tendo por base "df1"
        df = pd.merge(df1, df2, how='left', left_on='ID', right_on='ID')
        # Converte o float NaN para a string "NA"
        df['RSAUDE_ID'].replace(np.nan, 'NA', inplace=True)
        # Reordena as colunas priorizando as "mais" relevantes
        df = df[['ID', 'MUNNOME', 'MUNNOMEX', 'MUNCODDV', 'OBSERV', 'SITUACAO', 'MUNSINP',
                 'MUNSIAFI', 'UFCOD_ID', 'AMAZONIA', 'FRONTEIRA', 'CAPITAL', 'RSAUDE_ID',
                 'LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA', 'ANOINST', 'ANOEXT', 'SUCESSOR']]
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE', '?', '?', '?', '?', '?', '?', 'NA', '?',
                               '?', '?', 'NA', np.nan, np.nan, np.nan, np.nan, '?', '?', '?']
        return df


    # Função para adequar e formatar as colunas e valores da TCC TP_GESTAO (arquivo TP_GESTAO.cnv)
    def get_TP_GESTAO_treated(self):
        # Conversão da TCC TP_GESTAO para um objeto pandas DataFrame
        file_name = 'TP_GESTAO'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'GESTAO'}, inplace=True)
        # Considera da coluna GESTAO apenas a substring depois de duas letras maiúsculas e um espaço
        df['GESTAO'].replace(to_replace='^[A-Z]{2}\s{1}', value= '', regex=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC REGRA_C (arquivo REGRA_C.cnv)
    def get_REGRA_C_treated(self):
        # Conversão da TCC REGRA_C para um objeto pandas DataFrame
        file_name = 'REGRA_C'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'REGRA'}, inplace=True)
        # Modifica um valor da coluna ID
        df.loc['0', 'ID'] = '0000'
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC TP_ESTAB (arquivo TP_ESTAB.cnv)
    def get_TP_ESTAB_treated(self):
        # Conversão da TCC TP_ESTAB para um objeto pandas DataFrame
        file_name = 'TP_ESTAB'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'ESTABELECIMENTO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC ESFERA (arquivo ESFERA.cnv)
    def get_ESFERA_treated(self):
        # Conversão da TCC ESFERA para um objeto pandas DataFrame
        file_name = 'ESFERA'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'ESFERA'}, inplace=True)
        # Considera da coluna ESFERA apenas a substring depois de dois pontos
        df['ESFERA'].replace(to_replace='^\.\.', value= '', regex=True, inplace=True)
        # Modifica um valor da coluna ESFERA
        df.loc['0', 'ESFERA'] = 'ESTABELECIMENTO PRIVADO COM FINS LUCRATIVO PJ/PF'
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TB_SIGTAP (arquivo TB_SIGTAP.dbf)
    def get_TB_SIGTAP_treated(self):
        # Conversão da Tabela TB_SIGTAP para um objeto pandas DataFrame
        file_name = 'TB_SIGTAP'
        df = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df.rename(index=str, columns={'IP_COD': 'ID', 'IP_DSCR': 'PROCEDIMENTO'}, inplace=True)
        # Upload do arquivo "xlsx" que contém os PROC_SOLIC/PROC_REA presentes nos arquivos RDXXaamm (dos anos...
        # de 2008 a 2019) e não presentes no arquivo TB_SIGTAP.dbf. Ou seja,...
        # isso parece ser uma falha dos dados do Datasus
        dataframe = pd.read_excel(self.path + 'COD_PROCEDIMENTOS_OUT_TB_SIGTAP_ANOS_2008_2019' + '.xlsx')
        # Converte a coluna "ID" do objeto "dataframe" de "int" para "string"
        dataframe['ID'] = dataframe['ID'].astype('str')
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "dataframe" até...
        # formar uma "string" de tamanho = 10
        dataframe['ID'] = dataframe['ID'].apply(lambda x: x.zfill(10))
        # Adiciona a coluna DESCESTAB e respectivos valores ao objeto "dataframe"
        dataframe['PROCEDIMENTO'] = 'NAO PROVIDO NO ARQUIVO TB_SIGTAP.dbf'
        # Concatenação do objeto "dataframe" ao objeto "df"
        frames = []
        frames.append(df)
        frames.append(dataframe)
        dfinal = pd.concat(frames, ignore_index=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        dfinal.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena eventualmente as linhas por ordem crescente dos valores da coluna ID
        dfinal.sort_values(by=['ID'], inplace=True)
        # Reset eventualmente o index devido ao sorting prévio e à eventual eliminação de duplicates
        dfinal.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        dfinal.loc[dfinal.shape[0]] = ['NA', 'NOT AVAILABLE']
        return dfinal


    # Função para adequar e formatar as colunas e valores da TCC FINANC (arquivo FINANC.cnv)
    def get_FINANC_treated(self):
        # Conversão da TCC FINANC para um objeto pandas DataFrame
        file_name = 'FINANC'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'FINANCIAMENTO'}, inplace=True)
        # Considera da coluna GESTAO apenas a substring depois de duas letras maiúsculas e um espaço
        df['FINANCIAMENTO'].replace(to_replace='^\d{2}\s{1}', value= '', regex=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC COMPLEX (arquivo COMPLEX.cnv)
    def get_COMPLEX_treated(self):
        # Conversão da TCC COMPLEX para um objeto pandas DataFrame
        file_name = 'COMPLEX'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'COMPLEXIDADE'}, inplace=True)
        # Considera da coluna GESTAO apenas a substring depois de um dígito e um traço
        df['COMPLEXIDADE'].replace(to_replace='^\d{1}-', value= '', regex=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC DOCORIG (arquivo DOCORIG.cnv)
    def get_DOCORIG_treated(self):
        # Conversão da TCC DOCORIG para um objeto pandas DataFrame
        file_name = 'DOCORIG'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'TIPO_SIA'}, inplace=True)
        # Considera da coluna TIPO_SIA apenas a substring depois de dois pontos
        df['TIPO_SIA'].replace(to_replace='^\.\.', value= '', regex=True, inplace=True)
        # Modifica um valor da coluna TIPO_SIA
        df.loc['5', 'TIPO_SIA'] = 'APAC - PROCEDIMENTO SECUNDÁRIO'
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela CBO (arquivo CBO.dbf)
    def get_CBO_treated(self):
        # Conversão da Tabela CBO para um objeto pandas DataFrame
        file_name = 'CBO'
        df = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df.rename(index=str, columns={'CBO': 'ID', 'DS_CBO': 'OCUPACAO'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" até...
        # formar uma "string" de tamanho = 6
        df['ID'] = df['ID'].apply(lambda x: x.zfill(6))
        # Drop a linha inteira em que a coluna "ID" tem o valor especificado por representar código errado
        df = df.drop(df[df['ID']=='006540'].index)
        # Elimina linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        df.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena as linhas por ordem crescente dos valores da coluna ID
        df.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à eventual eliminação de duplicates
        df.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC MOTSAIPE (arquivo MOTSAIPE.cnv)
    def get_MOTSAIPE_treated(self):
        # Conversão da TCC MOTSAIPE para um objeto pandas DataFrame
        file_name = 'MOTSAIPE'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'MOTIVO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela S_CID (arquivo S_CID.dbf).
    def get_S_CID_treated(self):
        # Conversão da Tabela S_CID para um objeto pandas DataFrame
        file_name = 'S_CID'
        df = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df.rename(index=str, columns={'CD_COD': 'ID', 'CD_DESCR': 'DIAGNOSTICO'}, inplace=True)
        # Remove colunas indesejáveis do objeto pandas DataFrame
        df = df.drop(['OPC', 'CAT', 'SUBCAT', 'RESTRSEXO', 'CAMPOS_RAD', 'ESTADIO', 'REPETE_RAD'], axis=1)
        # Drop a linha inteira em que a coluna "ID" tem os valores especificados por não representarem valor útil
        df = df.drop(df[df['ID']=='0000'].index)
        df = df.drop(df[df['ID']=='9999'].index)
        ## Drop a linha inteira em que a coluna "DIAGNOSTICO" tem os valores especificados por não...
        # representarem valor útil
        df = df.drop(df[df['DIAGNOSTICO']=='CID NÇO IDENTIFICADO'].index)
        # Coloca todas as string da coluna especificada como UPPER CASE
        df['DIAGNOSTICO'] = df['DIAGNOSTICO'].apply(lambda x: x.upper())
        # Ordena as linhas de "df" por ordem crescente dos valores da coluna ID
        df.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio
        df.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC CARAT_AT (arquivo CARAT_AT.cnv)
    def get_CARAT_AT_treated(self):
        # Conversão da TCC CARAT_AT para um objeto pandas DataFrame
        file_name = 'CARAT_AT'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'CARATER'}, inplace=True)
        # Drop a linha inteira em que a coluna "ID" tem os valores especificados por não...
        # representarem valor útil
        df = df.drop(df[df['ID']=='0'].index)
        df = df.drop(df[df['ID']=='7'].index)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC FL_IDADE (arquivo FL_IDADE.cnv)
    def get_FL_IDADE_treated(self):
        # Conversão da TCC FL_IDADE para um objeto pandas DataFrame
        file_name = 'FL_IDADE'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'COMPATIBILIDADE'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC SEXO (arquivo SEXO.cnv)
    def get_SEXO_treated(self):
        # Conversão da TCC SEXO para um objeto pandas DataFrame
        file_name = 'SEXO'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'SEXO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC RACA_COR (arquivo RACA_COR.cnv)
    def get_RACA_COR_treated(self):
        # Conversão da TCC RACA_COR para um objeto pandas DataFrame
        file_name = 'RACA_COR'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'RACA'}, inplace=True)
        # Drop a linha inteira em que a coluna "ID" tem os valores especificados por não representarem valor útil
        df = df.drop(df[df['ID']=='00'].index)
        df = df.drop(df[df['ID']=='99'].index)
        df = df.drop(df[df['ID']=='06'].index)
        df = df.drop(df[df['ID']=='09'].index)
        df = df.drop(df[df['ID']=='1M'].index)
        # Considera da coluna ID apenas a substring depois um dígito
        df['ID'].replace(to_replace='^\d{1}', value= '', regex=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC INDICA (arquivo INDICA.cnv)
    def get_INDICA_treated(self):
        # Conversão da TCC INDICA para um objeto pandas DataFrame
        file_name = 'INDICA'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'SITUACAO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC CODOCO (arquivo CODOCO.cnv)
    def get_CODOCO_treated(self):
        # Conversão da TCC CODOCO para um objeto pandas DataFrame
        file_name = 'CODOCO'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'OCORRENCIA'}, inplace=True)
        # Modifica valores da coluna OCORRENCIA tal como previsto na TCC CODOCO
        df.loc['0', 'OCORRENCIA'] = '..APROVADO TOTALMENTE       (K)'
        df.loc['3', 'OCORRENCIA'] = '..ULTRAPASSOU TETO FISICO   (L)'
        df.loc['6', 'OCORRENCIA'] = '..PROCEDIMENTO SEM ORCAMENTO      (P)'
        # Substitui subtrings da coluna OCORRENCIA dependendo do valor da coluna ID
        df.loc[df.ID.str.startswith('1'), 'OCORRENCIA'] = \
        df.loc[df.ID.str.startswith('1'), 'OCORRENCIA'].replace(to_replace='^\.\.',
                                                                value= 'PRODUÇÃO TOT. APROVADA - ',
                                                                regex=True)
        df.loc[df.ID == '2L', 'OCORRENCIA'] = \
        df.loc[df.ID == '2L', 'OCORRENCIA'].replace(to_replace='^\.\.',
                                                    value= 'PRODUÇÃO PARC. APROVADA - ',
                                                    regex=True)
        df.loc[df.ID.str.startswith('3'), 'OCORRENCIA'] = \
        df.loc[df.ID.str.startswith('3'), 'OCORRENCIA'].replace(to_replace='^\.\.',
                                                                value= 'PRODUÇÃO PARC. APROVADA - ',
                                                                regex=True)
        df.loc[df.ID.str.startswith('4'), 'OCORRENCIA'] = \
        df.loc[df.ID.str.startswith('4'), 'OCORRENCIA'].replace(to_replace='^\.\.',
                                                                value= 'PRODUÇÃO NAO APROVADA - ',
                                                                regex=True)
        df.loc[df.ID == '5O', 'OCORRENCIA'] = \
        df.loc[df.ID == '5O', 'OCORRENCIA'].replace(to_replace='^\.\.',
                                                    value= 'PRODUÇÃO NAO APROVADA - ',
                                                    regex=True)
        # Desconsidera da coluna OCORRENCIA valores que terminam com uma substring da forma "(\.)"
        df['OCORRENCIA'].replace(to_replace='\(.\)$', value= '', regex=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC etnia (arquivo etnia.cnv)
    def get_etnia_treated(self):
        # Conversão da TCC etnia para um objeto pandas DataFrame
        file_name = 'etnia'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'INDIGENA'}, inplace=True)
        # Preenche os valores da coluna ID com zeros a esquerda até formar quatro digitos
        df['ID'] = df['ID'].apply(lambda x: x.zfill(4))
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela INE_EQUIPE_BR (arquivo INE_EQUIPE_BR.dbf)
    def get_INE_EQUIPE_BR_treated(self):
        # Conversão da Tabela INE_EQUIPE_BR para um objeto pandas DataFrame
        file_name = 'INE_EQUIPE_BR'
        df = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df.rename(index=str, columns={'CO_INE': 'ID', 'NO_REF': 'EQUIPE'}, inplace=True)
        # Considera da coluna EQUIPE apenas a substring depois de dez dígitos e um espaço
        df['EQUIPE'].replace(to_replace='^\d{10} ', value= '', regex=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC natjur (arquivo natjur.cnv)
    def get_natjur_treated(self):
        # Conversão da TCC natjur para um objeto pandas DataFrame
        file_name = 'natjur'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'NATUREZA'}, inplace=True)
        # Drop a linha inteira em que a coluna "ID" tem o valor especificado por não representar nada
        df = df.drop(df[df['ID']=='0'].index)
        # Reset o index devido à eliminação de linha efetuada no passo anterior
        df.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TABUF (arquivo TABUF.dbf)
    def get_TABUF_treated(self):
        # Conversão da Tabela TABUF para um objeto pandas DataFrame
        file_name = 'TABUF'
        df = download_table_dbf(file_name)
        # Renomeia colunas especificadas
        df.rename(index=str, columns={'CODIGO': 'ID', 'DESCRICAO': 'ESTADO'}, inplace=True)
        # Reordena as colunas
        df = df[['ID', 'ESTADO', 'SIGLA_UF']]
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE', '?']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela RSAUDE (do IBGE)
    def get_RSAUDE_treated(self):
        # Conversão da Tabela RSAUDE (em formato "xlsx") para um objeto pandas DataFrame
        df = pd.read_excel(self.path + 'RSAUDE' + '.xlsx')
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'REGIAO'}, inplace=True)
        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TB_GRUPO (arquivo TB_GRUPO.dbf)
    def get_TB_GRUPO_treated(self):
        # Conversão da Tabela TB_GRUPO para um objeto pandas DataFrame
        file_name = 'TB_GRUPO'
        df = download_table_dbf(file_name)
        # Renomeia colunas especificadas
        df.rename(index=str, columns={'CO_GRUPO': 'ID', 'NO_GRUPO': 'GRUPO'}, inplace=True)
        # Remove coluna indesejáveL do objeto pandas DataFrame
        df = df.drop(['DT_COMPET'], axis=1)
        # Torna UPPERCASE os valores da coluna GRUPO
        df['GRUPO'] = df['GRUPO'].str.upper()
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TB_SUBGR (arquivo TB_SUBGR.dbf)
    def get_TB_SUBGR_treated(self):
        # Conversão da Tabela TB_SUBGR para um objeto pandas DataFrame
        file_name = 'TB_SUBGR'
        df = download_table_dbf(file_name)
        # Renomeia colunas especificadas
        df.rename(index=str, columns={'CO_SUB_GRU': 'ID', 'NO_SUB_GRU': 'SUBGRUPO'}, inplace=True)
        # Remove coluna indesejáveL do objeto pandas DataFrame
        df = df.drop(['DT_COMPET'], axis=1)
        # Torna UPPERCASE os valores da coluna SUBGRUPO
        df['SUBGRUPO'] = df['SUBGRUPO'].str.upper()
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TB_FORMA (arquivo TB_FORMA.dbf)
    def get_TB_FORMA_treated(self):
        # Conversão da Tabela TB_FORMA para um objeto pandas DataFrame
        file_name = 'TB_FORMA'
        df = download_table_dbf(file_name)
        # Renomeia colunas especificadas
        df.rename(index=str, columns={'CO_FORMA': 'ID', 'NO_FORMA': 'FORMA'}, inplace=True)
        # Remove coluna indesejáveL do objeto pandas DataFrame
        df = df.drop(['DT_COMPET'], axis=1)
        # Torna UPPERCASE os valores da coluna FORMA
        df['FORMA'] = df['FORMA'].str.upper()
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df



if __name__ == '__main__':

    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)

    # Ajustar "the_path" para a localização dos arquivos "xlsx"
    the_path = os.getcwd()[:-len('\\transform')] + '\\files\\SIA\\'

    instancia = DataSiaAuxiliary(the_path)



###########################################################################################################################
# ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT ROOT #
###########################################################################################################################

import os

import psycopg2
from sqlalchemy import create_engine
import numpy as np
import pandas as pd

from utilities.essential_postgreSQL import files_in_ftp_subbase, get_tables_counts_subdb, files_loaded, files_to_load


def load_any(db_name, db_user, db_password):
    """
    Cria um schema do banco de dados "db_name" no SGBD PostgreSQL sem integridade referencial (de "primary and foreign
    keys") e insere nele dados das bases de dados (do Datasus) CNES, SIH, SIA, SINAN ou XXX no SGBD PostgreSQL pelo
    método copy_expert da classe cursor do pacote psycopg. Os dados são inseridos por sub-base de dados: CNES_ST, CNES_DC,
    CNES_PF, CNES_LT, CNES_EQ, CNES_SR, CNES_EP, CNES_HB, CNES_RC, CNES_GM, CNES_EE, CNES_EF, CNES_IN, SIH_RD, SIH_SP,
    SIA_PA, SINAN_DENG ou XXX. Assim, para cada sub-base de dados do Datasus é criado um schema.

    A inserção de dados consiste dos arquivos principais de dados em formato "dbc" e dos arquivos auxiliares de dados
    em formato "dbf", "cnv" e "xlsx" das sub-bases de dados CNES_ST (STXXaamm), CNES_DC (DCXXaamm), CNES_PF (PFXXaamm),
    CNES_LT (LTXXaamm), CNES_EQ (EQXXaamm), CNES_SR (SRXXaamm), CNES_EP (EPXXaamm), CNES_HB (HBXXaamm), CNES_RC (RCXXaamm),
    CNES_GM (GMXXaamm), CNES_EE (EEXXaamm), CNES_EF (EFXXaamm), CNES_IN (INXXaamm), SIH_RD (RDXXaamm), SIH_SP (SPXXaamm),
    SIA_PA (PAXXaamm), SINAN_DENG (DENGXXaa) ou XXX. Destaca-se que algumas bases de dados, como a do CNES, se subdividem
    em várias tabelas principais de dados, que no caso são em número de 13, conforme se pode contabilizar acima.

    Os arquivos principais de dados formam a tabela principal (child table) das respectiva sub-base de dados e estão em
    pastas específicas do endereço ftp do Datasus (ftp://ftp.datasus.gov.br/dissemin/publicos/) em formato "dbc". Cada
    arquivo "dbc" é baixado em tempo de execução, descompactado para "dbf", lido como um objeto pandas DataFrame e, para
    evitar a repetição do download, é salvo numa pasta criada dinamicamente e denominada "datasus_content" no computador
    de execução deste módulo no formato "parquet" no caso de nova necessidade desse arquivo principal de dados.

    Os arquivos secundários de dados formam as tabelas relacionais (parent tables) à tabela principal e estão em formato
    "dbf", "cnv" ou "xlsx". Os arquivos "dbf" e "cnv", presentes em diretórios do endereço ftp do Datasus, são baixados
    e convertidos em tempo de execução para objetos pandas DataFrame enquanto os arquivos "xlsx", quando necessários,
    foram  criados a partir de relações descritas no Dicionário de Dados da respectiva base de dado do Datasus e não
    retratadas em arquivos  "dbf" ou "cnv" ou a partir da incompletude de arquivos "dbf" ou "cnv".

    O referido diretório "datasus_content" onde são baixados os arquivos principais de dados em formato "parquet" pode
    ser alterado editando o módulo "folder" contido no subpackage "pegasus.dados.transform.extract".
    """

    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)

    print('\n*******************************************************************')
    print('O banco de dados PostgreSQL a ser preenchido deve antes ser criado!'.upper())
    print('*******************************************************************\n')

    decision = input("Digite 'q' se o banco de dados ainda não foi criado ou qualquer outra tecla em caso contrário: ")
    if decision == 'q':
        quit()
    else:
        # Nome do banco de dados do Datasus almejado
        datasus_db = input('\nDigite o nome da base de dados do Datasus (CNES/SIH/SIA/SINAN/XXX): ').lower()

        # Construção do objeto string do nome do sub-banco de dados do CNES, SIH, SIA, SINAN ou XXX
        # CNES
        if datasus_db == 'cnes':
            print('\nST: Estabelecimentos')
            print('DC: Dados Complementares')
            print('PF: Profissionais')
            print('LT: Leitos')
            print('EQ: Equipamentos')
            print('SR: Serviços Especializados')
            print('EP: Equipes')
            print('HB: Habilitações')
            print('GM: Gestão e Metas')
            print('EE: Estabelecimentos de Ensino')
            print('EF: Estabelecimentos Filantrópicos')
            print('IN: Incentivos')
            print('RC: Regras Contratuais')
            sub_db = input('\nDigite as iniciais da sub-base de dados do CNES (ST/DC/PF/LT/EQ/SR/EP/HB/RC/GM/EE/EF/IN): ').lower()
        # SIH
        elif datasus_db == 'sih':
            print('\nRD: Autorização de Internação Hospitalar - Reduzidas')
            print('SP: Autorização de Internação Hospitalar - Serviços Profissionais')
            sub_db = input('\nDigite as iniciais da sub-base de dados do SIH (RD/SP): ').lower()
        # SIA
        elif datasus_db == 'sia':
            print('\nPA: Procedimentos Ambulatoriais')
            sub_db = input('Digite as iniciais da sub-base de dados do SIA (PA/XX/.../YY): ').lower()
        # SINAN
        elif datasus_db == 'sinan':
            print('\nDENG: Dengue e Chikungunya\n')
            sub_db = input('Digite as iniciais da sub-base de dados do SINAN (DENG/XXXX/.../YYYY): ').lower()
        datasus_db += '_' + sub_db
        print('\nTo its goal...\n')

        # Criação de objetos string do nome das duas funções de inserção de dados do "datasus_db" contidas no respectivo...
        # módulo do package "insertion"
        str_most = 'insert_into_most_' + datasus_db.upper() + '_tables'
        str_main = 'insert_into_main_table_and_arquivos'

        # Importação das duas funções de inserção de dados do "datasus_db" usando a função nativa "__import__"
        if datasus_db.startswith('sinan'):
            #
            module1 = __import__('insertion.insert_into_any_' + datasus_db.upper()[:-5],
                                 fromlist=[str_most, str_main],
                                 level=0)
        else:
            #
            module1 = __import__('insertion.insert_into_any_' + datasus_db.upper()[:-3],
                                 fromlist=[str_most, str_main],
                                 level=0)

        # Colocação do nome das duas funções de inserção de dados importadas nas variáveis "most_tables" e "main_table"
        most_tables = getattr(module1, str_most)
        main_table = getattr(module1, str_main)

        # Chama a função "files_in_ftp_subbase" contida no módulo "essential_postgreSQL" do package "utilities" tendo...
        # como parâmetro a variável "datasus_db"
        df_arquivos_ftp = files_in_ftp_subbase(datasus_db)
        os.remove('stuff_ftp_files.txt')

        # Dados de conexão 1 (portanto o DB_NAME já deve ter sido previamente criado com esses dados)
        DB_NAME = db_name       # De acordo com o usuário; Nome do banco de dados mãe
        DB_HOST = '127.0.0.1'
        DB_PORT = '5432'
        DB_USER = db_user       # De acordo com o usuário
        DB_PASS = db_password   # De acordo com o usuário
        # Criação de objeto ndarray que armazena os referidos objetos string
        DB_DADOS = np.array([DB_NAME, DB_HOST, DB_PORT, DB_USER, DB_PASS])

        try:
            # Conecta ao banco de dados mãe "DB_NAME" do SGBD PostgreSQL usando o módulo python "psycopg2"
            conn = psycopg2.connect(dbname=DB_NAME, host=DB_HOST, port=DB_PORT, user=DB_USER, password=DB_PASS)
            print(conn.get_dsn_parameters(),'\n')
            # Criação de um cursor da conexão tipo "psycopg2" referenciado à variável "cursor"
            cursor = conn.cursor()
            # Inicializa o schema denominado "datasus_db" no banco de dados mãe "DB_NAME"
            cursor.execute(f'CREATE SCHEMA IF NOT EXISTS {datasus_db};')
            conn.commit()
            # Chama a função "get_tables_counts_subdb" contida no módulo "essential_postgreSQL" do package "utilities"
            dict_tabelas_e_counts_pg = get_tables_counts_subdb(cursor, datasus_db)
            # Encerra o cursor
            cursor.close()
            # Encerra a conexão
            conn.close()
        except (Exception, psycopg2.Error) as error:
            print('Erro ao conectar ao banco de dados PostgreSQL.', error)

        # Importação da função "create_tables" de criação do schema do banco de dados "datasus_db" existente no respectivo
        # módulo do package "schemas" usando a função python "__import__"
        module2 = __import__('schemas.sql_' + datasus_db.upper() + '_postgreSQL',
                             fromlist=['create_tables'],
                             level=0)
        # Referenciação da função "create_tables" à variável deste módulo denominada "create_schema"
        create_schema = getattr(module2, 'create_tables')
        # Uso da função "create_schema" para a criação do schema do banco de dados "datasus_db"
        create_schema(DB_DADOS, datasus_db)

        # Dados de conexão 2 (para uso da função "create_engine" do SQLAlchemy)
        DB_TYPE = 'postgresql'
        DB_DRIVER = 'psycopg2'
        # URI do banco de dados mãe "DB_NAME" do SGBD PostgreSQL
        DATABASE_URI = '%s+%s://%s:%s@%s:%s/%s' % (DB_TYPE, DB_DRIVER, DB_USER, DB_PASS, DB_HOST, DB_PORT, DB_NAME)
        # Cria um "engine" para o banco de dados mãe "DB_NAME" usando a função "create_engine" do SQLAlchemy
        engine = create_engine(DATABASE_URI)

        # Chama a função "files_loaded" contida no módulo "essential_postgreSQL" do package "utilities"
        df_arquivos_pg = files_loaded(dict_tabelas_e_counts_pg, datasus_db, engine)
        print(df_arquivos_pg)

        # Chama a função "files_to_load" contida no módulo "essential_postgreSQL" do package "utilities"
        df_arqs_nao_carregados = files_to_load(df_arquivos_ftp, df_arquivos_pg)

        # Quantidade de arquivos de dados principal que falta carregar em "datasus_db"
        qtd_arqs_datasus = df_arqs_nao_carregados.shape[0]
        print(f'\nA quantidade de arquivos principais de dados do {datasus_db} que falta carregar no {DB_NAME}/PostgreSQL é {qtd_arqs_datasus}.\n')

        # Cria o objeto string "path_xlsx" do diretório onde estão os eventuais arquivos "xlsx" relativos a "datasus_db"
        CURR_DIR = os.path.dirname(os.path.abspath(__file__)).replace('\\', '/')
        if datasus_db.startswith('cnes'):
            path_xlsx = CURR_DIR + '/files/CNES/'
        elif datasus_db.startswith('sih'):
            path_xlsx = CURR_DIR + '/files/SIH/'
        elif datasus_db.startswith('sia'):
            path_xlsx = CURR_DIR + '/files/SIA/'
        elif datasus_db == 'sim':
            path_xlsx = CURR_DIR + '/files/SIM/'
        elif datasus_db == 'sinasc':
            path_xlsx = CURR_DIR + '/files/SINASC/'
        elif datasus_db.startswith('sinan'):
            path_xlsx = CURR_DIR + '/files/SINAN/'

        # Remove a única key do objeto dict "dict_tabelas_e_counts_pg" que não se refere às parent tables do "datasus_db"...
        # utilizando o método "pop" da class "dict"
        dict_tabelas_e_counts_pg.pop('arquivos', None)
        # Inicialização de contador do número de parent tables do "datasus_db" sem dados
        count = 0
        # Faz a contagem de quantas parent tables do "datasus_db" estão sem dados, se é que estão
        for table in dict_tabelas_e_counts_pg:
            if dict_tabelas_e_counts_pg[table] == 0:
                count += 1

        # Carrega dados de tabela(s) auxiliar(es) (parent table(s)) do "datasus_db" se não constar dados dela(s) no...
        # "DBNAME" do PostgreSQL
        if ((dict_tabelas_e_counts_pg == dict()) or (count > 0)):
            print(f'\nIniciando a inserção de dados auxiliares no banco de dados {datasus_db} do {DB_NAME}/PostgreSQL usando pandas...')
            # Chama a função "most_tables" para inserção das tabelas auxiliares (parent tables) no "datasus_db" pelo...
            # método pandas.to_sql
            most_tables(path_xlsx, engine, datasus_db)
            print(f'Finalizou a inserção de dados auxiliares no banco de dados {datasus_db} do {DB_NAME}/PostgreSQL.')

            # Remoção de pastas vazias que contiveram arquivos "dbf" e/ou "cnv"
            if datasus_db in np.array(['cnes_st', 'cnes_lt', 'cnes_eq', 'cnes_sr', 'cnes_ep', 'cnes_ee', 'cnes_ef', 'cnes_in']):
                os.remove('TAB_CNES.zip')
                os.rmdir('TAB_DBF')
                os.rmdir('cnv')
            elif datasus_db in np.array(['sih_rd', 'sih_sp']):
                os.remove('TAB_SIH.zip')
                os.rmdir('CNV')
                os.rmdir('DBF')
            elif datasus_db in np.array(['sia_pa']):
                os.remove('TAB_SIA.zip')
                os.remove('TAB_SIH.zip')
                os.rmdir('CNV')
                os.rmdir('TAB_DBF')
            elif datasus_db == 'sim':
                os.remove('OBITOS_CID10_TAB.ZIP') # e do arquivo CNESDO18.dbf
            elif datasus_db == 'sinasc':
                os.remove('NASC_NOV_TAB.zip')
                os.remove('OBITOS_CID10_TAB.ZIP') # e do arquivo CNESDN18.dbf
            elif datasus_db == 'sinan_deng':
                os.remove('TAB_SINAN.zip')
            os.remove('base_territorial.zip')

        print(f'\nIniciando a inserção de dados principais no banco de dados {datasus_db} do {DB_NAME}/PostgreSQL usando copy_expert...')
        # Carrega dados da tabela principal do "datasus_db" no PostgreSQL
        for i in range(qtd_arqs_datasus):
            # Chama a função "main_table" para a inserção de dados na tabela principal (child table) + respectivas informações...
            # na tabela arquivos usando os métodos copy_expert + pandas.to_sql
            main_table(df_arqs_nao_carregados.NOME[i], df_arqs_nao_carregados.DIRETORIO[i],
                       df_arqs_nao_carregados.DATA_INSERCAO_FTP[i], engine, datasus_db, DB_DADOS)


###########################################################################################################################
# MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN #
###########################################################################################################################

if __name__ == '__main__':

    load_any('dbsus4', 'ericc', 'teste')



###########################################################################################################################
# SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM SIM #
###########################################################################################################################

import os
from datetime import datetime

import numpy as np
import pandas as pd

from transform.extract.download_SIM import download_DOXXaaaa, download_table_dbf, download_table_cnv


"""
Módulo de limpeza/tratamento de dados do SIM.

"""

# Função para converter um "value" num certo "type" de objeto ou caso não seja possível utiliza o valor "default"
def tryconvert(value, default, type):
    try:
        return type(value)
    except (ValueError, TypeError):
        return default


# Classe de dados principais do SIM
class DataSimMain:

    # Construtor
    def __init__(self, state, year):
        self.state = state
        self.year = year


    # Método para ler como um objeto pandas DataFrame o arquivo principal de dados do SIM e adequar e formatar suas...
    # colunas e valores
    def get_DOXXaaaa_treated(self):
        # Lê o arquivo "dbc" ou "parquet", se já tiver sido baixado, como um objeto pandas DataFrame
        dataframe = download_DOXXaaaa(self.state, self.year)
        print(f'O número de linhas do arquivo DO{self.state}{self.year} é {dataframe.shape[0]}.')

        for coluna in dataframe.columns.values:
            dataframe[coluna] = dataframe[coluna].apply(lambda x: x if '\x00' not in x else '')

        # Colunas definidas como necessárias no objeto pandas DataFrame que incrementará a tabela dobr
        lista_columns = np.array(['NUMERODO', 'CODINST', 'TIPOBITO', 'DTOBITO', 'HORAOBITO', 'NUMSUS',
                                  'NATURAL',  'CODMUNNATU', 'DTNASC', 'IDADE', 'SEXO', 'RACACOR', 'ESTCIV',
                                  'ESC', 'ESC2010', 'OCUP', 'CODMUNRES', 'LOCOCOR', 'CODESTAB', 'CODMUNOCOR',
                                  'TPMORTEOCO', 'ASSISTMED', 'EXAME', 'CIRURGIA', 'NECROPSIA', 'LINHAA',
                                  'LINHAB', 'LINHAC', 'LINHAD', 'LINHAII', 'CAUSABAS', 'CRM', 'DTATESTADO',
                                  'CIRCOBITO', 'ACIDTRAB', 'FONTE', 'TPPOS', 'DTINVESTIG', 'CAUSABAS_O',
                                  'DTCADASTRO', 'ATESTANTE', 'FONTEINV', 'DTRECEBIM', 'ATESTADO', 'ESCMAEAGR1',
                                  'ESCFALAGR1', 'STDOEPIDEM', 'STDONOVA', 'DIFDATA', 'DTCADINV', 'TPOBITOCOR',
                                  'DTCONINV', 'FONTES'])

        # Criação de um objeto pandas DataFrame vazio com as colunas especificadas acima
        df = pd.DataFrame(columns=lista_columns)

        # Colocação dos dados da variável "dataframe" na variável "df" nas colunas de mesmo nome preenchendo...
        # automaticamente com o float NaN as colunas da variável "df" não presentes na variável dataframe
        for col in df.columns.values:
            for coluna in dataframe.columns.values:
                if coluna == col:
                    df[col] = dataframe[coluna].tolist()
                    break

        # Coloca na variável "dif_set" o objeto array dos nomes das colunas da variável "df" que não estão...
        # presentes na variável "dataframe"
        dif_set = np.setdiff1d(df.columns.values, dataframe.columns.values)

        # Substitui o float NaN pela string vazia as colunas da variável "df" não presentes na variável "dataframe"
        for col in dif_set:
            df[col].replace(np.nan, '', inplace=True)

        # Exclui o último dígito numérico das colunas identificadas, o qual corresponde ao dígito de controle do...
        # código do município
        # Foi detectado que para alguns municípios o cálculo do dígito de controle não é válido
        # Esse dígito de controle esteve presente nos arquivos DOXXxxxx até o ano de 2005 (a confirmar!)
        if len(df.loc[0, 'CODMUNNATU']) == 7:
            df['CODMUNNATU'].replace(regex='.$',value='', inplace=True)
        if len(df.loc[0, 'CODMUNRES']) == 7:
            df['CODMUNRES'].replace(regex='.$',value='', inplace=True)
        if len(df.loc[0, 'CODMUNOCOR']) == 7:
            df['CODMUNOCOR'].replace(regex='.$',value='', inplace=True)

        # Simplifica/corrige a apresentação dos dados das colunas especificadas
        df['HORAOBITO'] = df['HORAOBITO'].apply(lambda x: x[:4] if len(x) > 4 else x)

        df['NATURAL'] = df['NATURAL'].apply(lambda x: x.zfill(3))

        df['OCUP'] = df['OCUP'].apply(lambda x: x.zfill(6))
        df['OCUP'] = df['OCUP'].apply(str.strip)
        df['OCUP'] = df['OCUP'].apply(lambda x: x if len(x) == 6 else '')

        df['CODESTAB'] = df['CODESTAB'].apply(lambda x: x.zfill(7))

        for col in np.array(['ESCMAEAGR1', 'ESCFALAGR1']):
            for i in np.array(['00', '01', '02', '03', '04', '05', '06', '07', '08', '09']):
                df[col].replace(i, str(int(i)), inplace=True)

        # Atualiza/corrige os labels das colunas especificadas
        df['NATURAL'].replace(['000', '999'], '', inplace=True)
        df['NATURAL'].replace('800', '001', inplace=True)
        df['NATURAL'].replace(['00.', '8s9'], '', inplace=True)

        for col in np.array(['DTOBITO', 'DTNASC']):
            df[col] = df[col].apply(lambda x: x if len(x) == 8 else '')
            df[col] = df[col].apply(lambda x: x if ' ' not in x else '')
            df[col] = df[col].apply(lambda x: x if '/' not in x else '')
            df[col] = df[col].apply(lambda x: x if '¾' not in x else '')
            df[col] = df[col].apply(lambda x: x if 'ó' not in x else '')
            df[col] = df[col].apply(lambda x: x if 1 <= tryconvert(x[0:2], 0, int) <= 31 else '')
            df[col] = df[col].apply(lambda x: x if 1 <= tryconvert(x[2:4], 0, int) <= 12 else '')

        for col in np.array(['CODMUNNATU', 'CODMUNRES', 'CODMUNOCOR']):
            df[col].replace(['000000', '150475', '421265', '422000', '431454',
                             '500627', '990002', '990010', '990014', '999999'], '', inplace=True)
            df[col].replace([str(i) for i in range(334501, 334531)], '330455', inplace=True)
            df[col].replace([str(i) for i in range(358001, 358059)], '355030', inplace=True)
            df[col].replace(['530000', '530500', '530600', '530800', '530900', '531700', '539901',
                             '539902', '539904', '539905', '539906', '539907', '539914', '539916',
                             '539918', '539919', '539920', '539921', '539924', '539925'], '530010', inplace=True)

        df['SEXO'].replace('1', 'M', inplace=True) # Label "M" de Masculino
        df['SEXO'].replace('2', 'F', inplace=True) # Label "F" de Feminino
        df['SEXO'].replace('0', '3', inplace=True)
        df['SEXO'].replace('3', 'IN', inplace=True) # Label "IN" de INdefinido

        df['ESTCIV'].replace(['²', '±'], '', inplace=True)

        df['ESC'].replace('A', '', inplace=True)

        df['OCUP'] = df['OCUP'].apply(lambda x: x if ' ' not in x else '')
        df['OCUP'] = df['OCUP'].apply(lambda x: x if '.' not in x else '')
        df['OCUP'] = df['OCUP'].apply(lambda x: x if '+' not in x else '')
        df['OCUP'] = df['OCUP'].apply(lambda x: x if 'X' not in x else '')
        df['OCUP'].replace('000000', '', inplace=True)

        df['CODESTAB'].replace('0000000', '', inplace=True)
        df['CODESTAB'].replace('2306840', '2461234', inplace=True)
        df['CODESTAB'].replace('2464276', '2726688', inplace=True)
        df['CODESTAB'].replace('2517825', '3563308', inplace=True)
        df['CODESTAB'].replace('2772299', '2465140', inplace=True)
        df['CODESTAB'].replace('3064115', '3401928', inplace=True)

        df['TPMORTEOCO'].replace('8', '6', inplace=True)
        df['TPMORTEOCO'].replace('9', '7', inplace=True)

        for col in np.array(['ASSISTMED', 'EXAME', 'CIRURGIA', 'NECROPSIA', 'ACIDTRAB']):
            df[col].replace(['0', '3', '4', '5', '6', '7', '8', '9'], '', inplace=True)
            df[col].replace('2', '0', inplace=True) # "2", representativo de "Não", é convertido para o objeto...
                                                    # string "0" do domínio binário

        for col in np.array(['CAUSABAS', 'CAUSABAS_O']):
            df[col].replace('B501', 'B508', inplace=True)
            df[col].replace('B656', 'B653', inplace=True)
            df[col].replace('C141', 'C140', inplace=True)
            df[col].replace('M723', 'M724', inplace=True)
            df[col].replace('M725', 'M728', inplace=True)
            df[col].replace('N975', 'N978', inplace=True)
            df[col].replace('Q314', 'P288', inplace=True)
            df[col].replace('Q350', 'Q351', inplace=True)
            df[col].replace('Q352', 'Q353', inplace=True)
            df[col].replace('Q354', 'Q355', inplace=True)
            df[col].replace(['Q356', 'Q358'], 'Q359', inplace=True)
            df[col].replace('R500', 'R508', inplace=True)
            df[col].replace('R501', 'R500', inplace=True)
            df[col].replace(['X590', 'X591', 'X592', 'X593', 'X594',
                             'X595', 'X596', 'X597', 'X598'], 'X599', inplace=True)
            df[col].replace('Y34', 'Y349', inplace=True)
            df[col].replace('Y447', 'Y448', inplace=True)

        df['CAUSABAS_O'].replace(regex='.$',value='', inplace=True)

        df['TPPOS'].replace('2', '0', inplace=True) # "2", representativo de "Não", é convertido para o objeto...
                                                    # string "0" do domínio binário

        df['DTATESTADO'].replace('09201608', '', inplace=True)
        df['DTATESTADO'] = df['DTATESTADO'].apply(lambda x: x if len(x) == 8 else '')
        df['DTATESTADO'] = df['DTATESTADO'].apply(lambda x: x if x[2:4] != '20' else '')

        df['CIRCOBITO'].replace(['á', 'ß', 'C'], '', inplace=True)

        for col in np.array(['ESCMAEAGR1', 'ESCFALAGR1']):
            df[col].replace('9', '', inplace=True)

        df['TPOBITOCOR'].replace('0', '', inplace=True)

        # Atribui um único label para uma mesma significação nas colunas especificadas
        df['TIPOBITO'].replace(['0', '3', '4', '5', '6', '7', '8' '9'], '', inplace=True)

        for col in np.array(['RACACOR', 'ESTCIV', 'ESC', 'LOCOCOR', 'ATESTANTE']):
            df[col].replace(['0', '6', '7', '8', '9'], '', inplace=True)

        df['ESC2010'].replace(['6', '7', '8', '9'], '', inplace=True)

        df['TPMORTEOCO'].replace(['0', '7', '8', '9'], '', inplace=True)

        for col in np.array(['CIRCOBITO', 'FONTE']):
            df[col].replace(['0', '5', '6', '7', '8', '9'], '', inplace=True)

        df['FONTEINV'].replace(['0', '9'], '', inplace=True)

        # Substitui uma string vazia pela string "NA" nas colunas de foreign keys
        # A coluna FONTES é apenas considerada como tal pois recebe tratamento específico mais adiante
        for col in np.array(['TIPOBITO', 'NATURAL', 'CODMUNNATU', 'RACACOR',
                             'ESTCIV', 'ESC', 'ESC2010', 'OCUP', 'CODMUNRES',
                             'LOCOCOR', 'CODESTAB', 'CODMUNOCOR', 'TPMORTEOCO',
                             'CAUSABAS', 'CIRCOBITO', 'FONTE', 'CAUSABAS_O',
                             'ATESTANTE', 'FONTEINV', 'ESCMAEAGR1', 'ESCFALAGR1',
                             'TPOBITOCOR', 'FONTES']):
            df[col].replace('', 'NA', inplace=True)

        # Substitui uma string vazia por None nas colunas de atributos especificadas
        for col in np.array(['CODINST', 'HORAOBITO', 'NUMSUS', 'SEXO', 'LINHAA',
                             'LINHAB', 'LINHAC', 'LINHAD', 'LINHAII', 'CRM', 'ATESTADO']):
            df[col].replace('', None, inplace=True)

        # Divisão da coluna "FONTES" em seis colunas conforme Dicionário de Dados da Tabela DOM ("M" de...
        # investigação materna)
        df['FONTES'] = df['FONTES'].apply(lambda x: x if len(x) == 6 else x)
        for col in np.array(['FONTENTREV', 'FONTEAMBUL', 'FONTEPRONT', 'FONTESVO', 'FONTEIML', 'FONTEPROF']):
            df[col] = df['FONTES'].apply(lambda x: 'NA' if x == 'NA' else x[0]) # O valor quando a condição...
                                                                                # "else" se verifica é "S" de "Sim"
            df[col].replace('X', '0', inplace=True) # Substitui a string "X" por "0" de "Não" tornando a...
                                                    # coluna "col" com domínio "binário"
            df[col].replace('S', '1', inplace=True) # Substitui a string "X" por "1" de "Sim" tornando a...
                                                    # coluna "col" com domínio "binário"

        # Eliminação da coluna "FONTES" por se tornar desnecessária com a adição das seis colunas especificadas acima
        df.drop('FONTES', axis=1, inplace=True)
        # Converte do tipo string para datetime as colunas especificadas substituindo as datas faltantes ("NaT") pela...
        # data futura "2099-01-01" para permitir a inserção das referidas colunas no SGBD postgreSQL
        for col in np.array(['DTOBITO', 'DTNASC', 'DTATESTADO', 'DTINVESTIG',
                             'DTCADASTRO', 'DTRECEBIM', 'DTCADINV', 'DTCONINV']):
            df[col] = df[col].apply(lambda x: datetime.strptime(x, '%d%m%Y').date() \
                                    if x != '' else datetime(2099, 1, 1).date())

        # Verifica se as datas das colunas especificadas são absurdas e em caso afirmativo as substitui pela...
        # data futura "2099-01-01"
        for col in np.array(['DTOBITO', 'DTATESTADO', 'DTINVESTIG',
                             'DTCADASTRO', 'DTRECEBIM', 'DTCADINV', 'DTCONINV']):
            df[col] = df[col].apply(lambda x: x if datetime(2000, 12, 31).date() < x < \
                                    datetime(2020, 12, 31).date() else datetime(2099, 1, 1).date())

        df['DTNASC'] = df['DTNASC'].apply(lambda x: x if datetime(1850, 12, 31).date() < x < \
                                          datetime(2020, 12, 31).date() else datetime(2099, 1, 1).date())

        # Computa a diferença entre as datas de óbito e de nascimento em dias e a aloca como a coluna "IDADE"...
        # do objeto pandas DataFrame
        df['IDADE'] = df['DTOBITO'] - df['DTNASC']

        # Converte os valores da coluna IDADE de datetime.timedelta para string
        # Ainda na mesma linha, cria uma lista de dois objetos string de cada valor da coluna IDADE e aproveita...
        # apenas o primeiro objeto de cada lista
        df['IDADE'] = df['IDADE'].apply(lambda x: str(x).split(' day')[0])

        # Os valores em que a operação anterior forneceu a string "0:00:00" são substituídos pela string...
        # "0" (RN que viveram menos de um dia)
        df['IDADE'].replace('0:00:00', '0', inplace=True)

        # Converte os valores da coluna IDADE de string para float (em dias) atribuindo o float NaN para as...
        # string que começam com "-"
        df['IDADE'] = df['IDADE'].apply(lambda x: np.nan if x[0] == '-' else float(x))

        # Transforma o valor da coluna referida de dias para anos mantendo cinco casas decimais
        df['IDADE']=df['IDADE'].div(365).round(5)

        # Converte do tipo object para int ou para None as colunas de atributos de valores binários (0 ou 1)
        for col in np.array(['ASSISTMED', 'EXAME', 'CIRURGIA', 'NECROPSIA', 'ACIDTRAB',
                             'TPPOS', 'STDOEPIDEM', 'STDONOVA', 'FONTENTREV', 'FONTEAMBUL',
                             'FONTEPRONT', 'FONTESVO', 'FONTEIML', 'FONTEPROF']):
            df[col] = df[col].apply(lambda x: tryconvert(x, None, int))

        # Converte do tipo object para float sem casas decimais as colunas de atributos de valores...
        # representativos de quantidades ou para o valor None caso a coluna esteja com a string vazia
        df['DIFDATA'] = df['DIFDATA'].apply(lambda x: round(float(x),0) if x != '' else None)

        # Reordena o objeto pandas DataFrame por ordem crescente de valores da coluna DTOBITO
        df.sort_values(by=['DTOBITO'], inplace=True)

        # Renumera os índices devido à alteração efetivada no passo anterior
        df.reset_index(drop=True, inplace=True)

        # Renomeia colunas
        df.rename(index=str, columns={'TIPOBITO': 'TIPOBITO_ID', 'NATURAL': 'NATURALE_ID',
                                      'CODMUNNATU': 'CODMUNNATU_ID', 'RACACOR': 'RACACOR_ID',
                                      'ESTCIV': 'ESTCIV_ID', 'ESC': 'ESC_ID',
                                      'ESC2010': 'ESC2010_ID', 'OCUP': 'OCUP_ID',
                                      'CODMUNRES': 'CODMUNRES_ID', 'LOCOCOR': 'LOCOCOR_ID',
                                      'CODESTAB': 'CODESTAB_ID', 'CODMUNOCOR': 'CODMUNOCOR_ID',
                                      'TPMORTEOCO': 'TPMORTEOCO_ID', 'CAUSABAS': 'CAUSABAS_ID',
                                      'CIRCOBITO': 'CIRCOBITO_ID', 'FONTE': 'FONTE_ID',
                                      'CAUSABAS_O': 'CAUSABAS_O_ID', 'ATESTANTE': 'ATESTANTE_ID',
                                      'FONTEINV': 'FONTEINV_ID', 'ESCMAEAGR1': 'ESCMAEAGR1_ID',
                                      'ESCFALAGR1': 'ESCFALAGR1_ID', 'TPOBITOCOR': 'TPOBITOCOR_ID'}, inplace=True)

        print(f'Tratou o arquivo DO{self.state}{self.year} (shape final: {df.shape[0]} x {df.shape[1]}).')

        return df


# Classe de dados auxiliares do SIM
class DataSimAuxiliary:

    # Construtor
    def __init__(self, path):
        self.path = path


    # Função para adequar e formatar as colunas e valores da TCC TIPOBITO (arquivo TIPOBITO.cnv)
    def get_TIPOBITO_treated(self):
        # Conversão da TCC TIPOBITO para um objeto pandas DataFrame
        file_name = 'TIPOBITO'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'TIPO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC NAT1212 (arquivo NAT1212.cnv)
    def get_NAT1212_treated(self):
        # Conversão da TCC NAT1212 para um objeto pandas DataFrame
        file_name = 'NAT1212'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'LOCAL'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "df" até formar uma...
        # "string" de tamanho = 3
        df['ID'] = df['ID'].apply(lambda x: x.zfill(3))
        # Upload do arquivo "xlsx" que contém os NATURAL presentes nos arquivos DOXXxxxx (a partir do ano...
        # de 2001) e não presentes na TCC NAT1212. Ou seja, isso parece ser uma falha dos dados do Datasus
        dataframe = pd.read_excel(self.path + 'NATURAL_OUT_NAT1212_ANOS_1997_2017' + '.xlsx')
        # Converte a coluna "ID" do objeto "dataframe" de "int" para "string"
        dataframe['ID'] = dataframe['ID'].astype('str')
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "dataframe" até...
        # formar uma "string" de tamanho = 3
        dataframe['ID'] = dataframe['ID'].apply(lambda x: x.zfill(3))
        # Adiciona a coluna "LOCAL" e respectivos valores ao objeto "dataframe"
        dataframe['LOCAL'] = ['NAO PROVIDO NA TCC NAT1212'] * (dataframe.shape[0])
        # Concatenação do objeto "dataframe" ao objeto "df"
        frames = []
        frames.append(df)
        frames.append(dataframe)
        dfinal = pd.concat(frames, ignore_index=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        dfinal.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena eventualmente as linhas por ordem crescente dos valores da coluna ID
        dfinal.sort_values(by=['ID'], inplace=True)
        # Reset eventualmente o index devido ao sorting prévio e à eventual eliminação de duplicates
        dfinal.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar célula de...
        # string vazia da coluna "NATURAL_ID" da tabela DOBR
        dfinal.loc[dfinal.shape[0]] = ['NA', 'NOT AVAILABLE']
        return dfinal


    # Função para adequar e formatar as colunas e valores da Tabela TABUF (arquivo TABUF.dbf)
    def get_TABUF_treated(self):
        # Conversão da Tabela TABUF para um objeto pandas DataFrame
        file_name = 'TABUF'
        df = download_table_dbf(file_name)
        # Renomeia colunas especificadas
        df.rename(index=str, columns={'CODIGO': 'ID', 'DESCRICAO': 'ESTADO'}, inplace=True)
        # Reordena as colunas
        df = df[['ID', 'ESTADO', 'SIGLA_UF']]
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE', '?']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela RSAUDE (do IBGE)
    def get_RSAUDE_treated(self):
        # Conversão da Tabela RSAUDE (em formato "xlsx") para um objeto pandas DataFrame
        df = pd.read_excel(self.path + 'RSAUDE' + '.xlsx')
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'REGIAO'}, inplace=True)
        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Método para adequar e formatar as colunas e valores da Tabela CADMUN (arquivo CADMUN.dbf)
    def get_CADMUN_treated(self):
        # Conversão da Tabela CADMUN para um objeto pandas DataFrame
        file_name = 'CADMUN'
        df1 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df1.rename(index=str, columns={'MUNCOD': 'ID', 'UFCOD': 'UFCOD_ID'}, inplace=True)
        # Drop a linha inteira em que a coluna "ID" tem o valor especificado por não representar nenhum município
        df1 = df1.drop(df1[df1['ID']=='000000'].index)
        # Remove colunas indesejáveis do objeto pandas DataFrame
        df1 = df1.drop(['MUNSINON', 'MUNSINONDV', 'MESOCOD', 'MICROCOD', 'MSAUDCOD',
                        'RSAUDCOD', 'CSAUDCOD', 'RMETRCOD', 'AGLCOD'], axis=1)
        # Substitui uma string vazia pela string "?" nas colunas especificadas
        for col in ['SITUACAO', 'MUNSINP', 'MUNSIAFI', 'MUNNOME', 'MUNNOMEX', 'OBSERV',
                    'AMAZONIA', 'FRONTEIRA', 'CAPITAL', 'ANOINST', 'ANOEXT', 'SUCESSOR']:
            df1[col].replace('', '?', inplace=True)
        # Substitui uma string vazia pela string "NA" nas colunas especificadas
        df1['UFCOD_ID'].replace('', 'NA', inplace=True)
        # Substitui uma string vazia pelo float "NaN" nas colunas especificadas
        for col in ['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']:
            df1[col].replace('', np.nan, inplace=True)
        # Converte do tipo object para float as colunas especificadas
        df1[['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']] = \
        df1[['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']].astype('float')
        # Coloca todas as string das colunas especificadas como UPPER CASE
        df1['MUNNOME'] = df1['MUNNOME'].apply(lambda x: x.upper())
        df1['MUNNOMEX'] = df1['MUNNOMEX'].apply(lambda x: x.upper())
        # Insere uma linha referente ao Município de Nazária/PI não constante originalmente do arquivo
        df1.loc[df1.shape[0]] = ['220672', '2206720', 'ATIVO', '?', '?', 'NAZÁRIA', 'NAZARIA', '?',
                                 'N', 'N', 'N', '22', '?', '?', '?', np.nan, np.nan, np.nan, 363.589]
        # Ordena as linhas de "df1" por ordem crescente dos valores da coluna ID
        df1.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à exclusão e inclusão das linhas referidas acima
        df1.reset_index(drop=True, inplace=True)
        # Conversão da Tabela rl_municip_regsaud para um objeto pandas DataFrame
        file_name = 'rl_municip_regsaud'
        df2 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df2.rename(index=str, columns={'CO_MUNICIP': 'ID', 'CO_REGSAUD': 'RSAUDE_ID'}, inplace=True)
        # Faz o merge de "df1" e "df2" pela coluna ID tendo por base "df1"
        df = pd.merge(df1, df2, how='left', left_on='ID', right_on='ID')
        # Converte o float NaN para a string "NA"
        df['RSAUDE_ID'].replace(np.nan, 'NA', inplace=True)
        # Reordena as colunas priorizando as "mais" relevantes
        df = df[['ID', 'MUNNOME', 'MUNNOMEX', 'MUNCODDV', 'OBSERV', 'SITUACAO', 'MUNSINP',
                 'MUNSIAFI', 'UFCOD_ID', 'AMAZONIA', 'FRONTEIRA', 'CAPITAL', 'RSAUDE_ID',
                 'LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA', 'ANOINST', 'ANOEXT', 'SUCESSOR']]
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE', '?', '?', '?', '?', '?', '?', 'NA', '?',
                               '?', '?', 'NA', np.nan, np.nan, np.nan, np.nan, '?', '?', '?']
        return df


    # Função para adequar e formatar as colunas e valores da TCC RACA (arquivo RACA.cnv)
    def get_RACA_treated(self):
        # Conversão da TCC RACA para um objeto pandas DataFrame
        file_name = 'RACA'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'TIPO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df

    # Função para adequar e formatar as colunas e valores da TCC ESTCIV (arquivo ESTCIV.cnv)
    def get_ESTCIV_treated(self):
        # Conversão da TCC ESTCIV para um objeto pandas DataFrame
        file_name = 'ESTCIV'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'SITUACAO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC INSTRUC (arquivo INSTRUC.cnv)
    def get_INSTRUC_treated(self):
        # Conversão da TCC INSTRUC para um objeto pandas DataFrame
        file_name = 'INSTRUC'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'FAIXA_DE_ANOS_INSTRUCAO'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC ESCSERIE (arquivo ESCSERIE.cnv)
    def get_ESCSERIE_treated(self):
        # Conversão da TCC ESCSERIE para um objeto pandas DataFrame
        file_name = 'ESCSERIE'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'ESCOLARIDADE'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TABOCUP (arquivo TABOCUP.dbf)
    # e das TCC CBO2002 e OCUPA (arquivos CBO2002.cnv e OCUPA.cnv, respectivamente)
    def get_TABOCUP_2TCC_treated(self):
        # Conversão da Tabela TABOCUP para um objeto pandas DataFrame
        file_name = 'TABOCUP'
        df1 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df1.rename(index=str, columns={'CODIGO': 'ID', 'DESCRICAO': 'OCUPACAO'}, inplace=True)
        # Ordena as linhas de "df1" por ordem crescente dos valores da coluna ID
        df1.sort_values(by=['ID'], inplace=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        df1.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Reset o index devido ao sorting prévio
        df1.reset_index(drop=True, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "df1" até formar...
        # uma "string" de tamanho = 6
        df1['ID'] = df1['ID'].apply(lambda x: x.zfill(6))
        # Conversão da TCC CBO2002 para um objeto pandas DataFrame
        file_name = 'CBO2002'
        df2 = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df2.rename(index=str, columns={'SIGNIFICACAO': 'OCUPACAO'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "df2" até formar...
        # uma "string" de tamanho = 6
        df2['ID'] = df2['ID'].apply(lambda x: x.zfill(6))
        # Conversão da TCC OCUPA para um objeto pandas DataFrame
        file_name = 'OCUPA'
        df3 = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df3.rename(index=str, columns={'SIGNIFICACAO': 'OCUPACAO'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "df3" até formar...
        # uma "string" de tamanho = 6
        df3['ID'] = df3['ID'].apply(lambda x: x.zfill(6))
        # Concatena os três objetos pandas DataFrame
        frames = []
        frames.append(df1)
        frames.append(df2)
        frames.append(df3)
        df = pd.concat(frames, ignore_index=True)
        # Elimina linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        df.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena as linhas por ordem crescente dos valores da coluna "ID"
        df.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à eventual eliminação de duplicates
        df.reset_index(drop=True, inplace=True)
        # Upload do arquivo "xlsx" que contém os OCUP presentes nos arquivos DOXXaaaa (dos anos de...
        #  1997 a 2017) e não presentes na Tabela TABOCUP e nas TCC CBO2002 e OCUPA. Ou seja, isso...
        # parece ser uma falha dos dados do Datasus
        dataframe = pd.read_excel(self.path + 'OCUP_OUT_TABOCUP_E_2TCC_ANOS_1997_2017' + '.xlsx')
        # Converte a coluna "ID" do objeto "dataframe" de "int" para "string"
        dataframe['ID'] = dataframe['ID'].astype('str')
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "ID" do objeto "dataframe" até formar...
        # uma "string" de tamanho = 6
        dataframe['ID'] = dataframe['ID'].apply(lambda x: x.zfill(6))
        # Adiciona a coluna "OCUPACAO" e respectivos valores ao objeto "dataframe"
        dataframe['OCUPACAO'] = ['NAO PROVIDO EM TABOCUP.DBF E NAS TCC CBO2002/OCUPA'] * (dataframe.shape[0])
        # Concatenação do objeto "dataframe" ao objeto "df"
        frames = []
        frames.append(df)
        frames.append(dataframe)
        dfinal = pd.concat(frames, ignore_index=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        dfinal.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena eventualmente as linhas por ordem crescente dos valores da coluna ID
        dfinal.sort_values(by=['ID'], inplace=True)
        # Reset eventualmente o index devido ao sorting prévio e à eventual eliminação de duplicates
        dfinal.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        dfinal.loc[dfinal.shape[0]] = ['NA', 'NOT AVAILABLE']
        return dfinal


    # Função para adequar e formatar as colunas e valores da TCC LOCOCOR (arquivo LOCOCOR.cnv)
    def get_LOCOCOR_treated(self):
        # Conversão da TCC LOCOCOR para um objeto pandas DataFrame
        file_name = 'LOCOCOR'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'LUGAR'}, inplace=True)
        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela CNESDO18 do SIM (arquivo CNESDO18.dbf)...
    # e da TCC ESTAB06 (arquivo ESTAB06.cnv)
    # Além disso faz o "merge" a elas das TCC ESFERA e NAT_ORG (arquivos ESFERA.cnv e NAT_ORG.cnv, respectivamente)
    def get_CNESDO18_3TCC_treated(self):
        # Conversão da Tabela CNESDO18 para um objeto pandas DataFrame
        file_name = 'CNESDO18'
        df1 = download_table_dbf(file_name)
        # Ordena as linhas de "df1" por ordem crescente dos valores da coluna CODESTAB
        df1.sort_values(by=['CODESTAB'], inplace=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna CODESTAB e mantém a primeira ocorrência
        df1.drop_duplicates(subset='CODESTAB', keep='first', inplace=True)
        # Reset o index devido ao sorting prévio
        df1.reset_index(drop=True, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "CODESTAB" do objeto "df1" até formar...
        # uma "string" de tamanho = 7
        df1['CODESTAB'] = df1['CODESTAB'].apply(lambda x: x.zfill(7))
        # Conversão da TCC ESTAB06 para um objeto pandas DataFrame
        file_name = 'ESTAB06'
        df2 = download_table_cnv(file_name)
        df2.rename(index=str, columns={'ID': 'CODESTAB', 'SIGNIFICACAO': 'DESCESTAB'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "CODESTAB" do objeto "df2" até formar...
        # uma "string" de tamanho = 7
        df2['CODESTAB'] = df2['CODESTAB'].apply(lambda x: x.zfill(7))
        # Concatena os dois objetos pandas DataFrame
        frames = []
        frames.append(df1)
        frames.append(df2)
        df = pd.concat(frames, ignore_index=True)
        # Elimina linhas duplicadas
        df.drop_duplicates(subset='CODESTAB', keep='first', inplace=True)
        # Ordena as linhas por ordem crescente dos valores da coluna "CODESTAB"
        df.sort_values(by=['CODESTAB'], inplace=True)
        # Reseta os índices
        df.reset_index(drop=True, inplace=True)
        # Conversão da TCC ESFERA18 para um objeto pandas DataFrame
        file_name = 'ESFERA18'
        df3 = download_table_cnv(file_name)
        # Adequa e formata a TCC ESFERA18
        df3.rename(index=str, columns={'ID': 'CODESTAB', 'SIGNIFICACAO': 'ESFERA'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "CODESTAB" do objeto "df3" até formar...
        # uma "string" de tamanho = 7
        df3['CODESTAB'] = df3['CODESTAB'].apply(lambda x: x.zfill(7))
        # Conversão da TCC NAT_ORG (já em formato "xlsx" e não "cnv") para um objeto pandas DataFrame
        file_name = 'NAT_ORG'
        df4 = download_table_cnv(file_name)
        # Adequa e formata a TCC NAT_ORG
        df4.rename(index=str, columns={'ID': 'CODESTAB', 'SIGNIFICACAO': 'REGIME'}, inplace=True)
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "CODESTAB" do objeto "df4" até formar...
        # uma "string" de tamanho = 7
        df4['CODESTAB'] = df4['CODESTAB'].apply(lambda x: x.zfill(7))
        # Realiza o "merge" da TCC ESFERA18 à TCC NAT_ORG
        df5 = df3.append(df4, sort=False)
        df6 = df5.replace(np.nan,'').groupby('CODESTAB',as_index=False).agg(''.join)
        df6.sort_values(by=['CODESTAB'], inplace=True)
        df6.reset_index(drop=True, inplace=True)
        # Realiza o "merge" da TCC ESFERA18 (+ TCC NAT_ORG) à (Tabela CNESDO18 + TCC ESTAB06)
        df7 = df.append(df6, sort=False)
        df8 = df7.replace(np.nan,'').groupby('CODESTAB',as_index=False).agg(''.join)
        df8.sort_values(by=['CODESTAB'], inplace=True)
        df8.reset_index(drop=True, inplace=True)
        # Substitui os valores de string vazia das colunas especificadas pela string "?"
        df8['DESCESTAB'].replace('','?', inplace=True)
        df8['ESFERA'].replace('','?', inplace=True)
        df8['REGIME'].replace('','?', inplace=True)
        # Upload do arquivo "xlsx" que contém os CODESTAB presentes nos arquivos DOXXaaaa (dos anos de...
        # 1997 a 2017) e não presentes na tabela CNESDO18 e nas TCC ESTAB06, ESFERA18 e NAT_ORG. Ou seja,...
        # isso parece ser uma falha dos dados do Datasus
        dataframe = pd.read_excel(self.path + 'CODESTAB_OUT_CNESDO18_E_3TCC_ANOS_1997_2017' + '.xlsx')
        # Converte a coluna "CODESTAB" do objeto "dataframe" de "int" para "string"
        dataframe['CODESTAB'] = dataframe['CODESTAB'].astype('str')
        # Adiciona zeros à esquerda nos valores (tipo string) da coluna "CODESTAB" do objeto "dataframe" até formar...
        # uma "string" de tamanho = 7
        dataframe['CODESTAB'] = dataframe['CODESTAB'].apply(lambda x: x.zfill(7))
        # Adiciona as colunas "DESCESTAB", "ESFERA" e "REGIME" e respectivos valores ao objeto "dataframe"
        dataframe['DESCESTAB'] = ['NAO PROVIDO EM CNESDO18.DBF E NAS TCC ESTAB06/ESFERA18/NAT_ORG'] * (dataframe.shape[0])
        dataframe['ESFERA'] = ['?'] * (dataframe.shape[0])
        dataframe['REGIME'] = ['?'] * (dataframe.shape[0])
        # Concatenação do objeto "dataframe" ao objeto "df8"
        frames = []
        frames.append(df8)
        frames.append(dataframe)
        dfinal = pd.concat(frames, ignore_index=True)
        # Renomeia a coluna "CODESTAB"
        dfinal.rename(index=str, columns={'CODESTAB': 'ID'}, inplace=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        dfinal.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena eventualmente as linhas por ordem crescente dos valores da coluna ID
        dfinal.sort_values(by=['ID'], inplace=True)
        # Reset eventualmente o index devido ao sorting prévio e à eventual eliminação de duplicates
        dfinal.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        dfinal.loc[dfinal.shape[0]] = ['NA', 'NOT AVAILABLE', '?', '?']
        return dfinal


    # Função para adequar e formatar as colunas e valores da Tabela TPMORTEOCO ("constando" apenas...
    # do Dicionário de Dados do SIM)
    def get_TPMORTEOCO_treated(self):
        # Conversão da Tabela TPMORTEOCO (em formato "xlsx") para um objeto pandas DataFrame
        df = pd.read_excel(self.path + 'TPMORTEOCO' + '.xlsx')
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'EPOCA_MORTE'}, inplace=True)
        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela CID10 (arquivo CID10.dbf) e de...
    # 21 TCC com nome CID10_XX (arquivos "cnv") sendo XX indo de 01 a 21, um para cada capítulo do CID 10.
    def get_CID10_treated(self):
        # Conversão da Tabela CID10 para um objeto pandas DataFrame
        file_name = 'CID10'
        df1 = download_table_dbf(file_name)
        # Remove colunas indesejáveis do objeto pandas DataFrame
        df1 = df1.drop(['OPC', 'CAT', 'SUBCAT', 'RESTRSEXO'], axis=1)
        # Renomeia as colunas especificadas
        df1.rename(index=str, columns={'CID10': 'ID', 'DESCR': 'DOENCA'}, inplace=True)
        # Coloca todas as string da coluna especificada como UPPER CASE
        df1['DOENCA'] = df1['DOENCA'].apply(lambda x: x.upper())
        # Ordena as linhas de "df1" por ordem crescente dos valores da coluna ID
        df1.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio
        df1.reset_index(drop=True, inplace=True)
        # Conversão das 21 TCC CID10_XX para um objeto pandas DataFrame
        frames = []
        for i in range(1, 22):
            i = str(i).zfill(2)
            file_name = 'CID10_' + i
            dfi = download_table_cnv(file_name)
            frames.append(dfi)
        df2 = pd.concat(frames, ignore_index=True)
        df2.drop_duplicates(subset='ID', keep='first', inplace=True)
        df2.sort_values(by=['ID'], inplace=True)
        df2.reset_index(drop=True, inplace=True)
        # Renomeia a coluna SIGNIFICACAO
        df2.rename(index=str, columns={'SIGNIFICACAO': 'DOENCA'}, inplace=True)
        # Concatena os dois objetos pandas DataFrame
        frames = []
        frames.append(df1)
        frames.append(df2)
        df = pd.concat(frames, ignore_index=True)
        # Elimina linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        df.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena as linhas de "df" por ordem crescente dos valores da coluna ID
        df.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à eventual eliminação de duplicates
        df.reset_index(drop=True, inplace=True)
        # Upload do arquivo "xlsx" que contém os CAUSABAS OU CAUSABAS_O presentes nos arquivos...
        # DOXXaaaa (dos anos de 1997 a 2017) e não presentes na Tabela CID10 ou nas TCC CID10. Ou seja,...
        # isso parece ser uma falha dos dados do Datasus
        dataframe = pd.read_excel(self.path + 'CAUSABAS_OUT_CID10_ANOS_1997_2017' + '.xlsx')
        # Adiciona a coluna "DOENCA" e respectivos valores ao objeto "dataframe"
        dataframe['DOENCA'] = ['NAO PROVIDO EM CID10.DBF E NAS TCC CID10'] * (dataframe.shape[0])
        # Concatenação do objeto "dataframe" ao objeto "df"
        frames = []
        frames.append(df)
        frames.append(dataframe)
        dfinal = pd.concat(frames, ignore_index=True)
        # Elimina eventuais linhas duplicadas tendo por base a coluna ID e mantém a primeira ocorrência
        dfinal.drop_duplicates(subset='ID', keep='first', inplace=True)
        # Ordena eventualmente as linhas por ordem crescente dos valores da coluna ID
        dfinal.sort_values(by=['ID'], inplace=True)
        # Reset eventualmente o index devido ao sorting prévio e à eventual eliminação de duplicates
        dfinal.reset_index(drop=True, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        dfinal.loc[dfinal.shape[0]] = ['NA', 'NOT AVAILABLE']
        return dfinal


    # Função para adequar e formatar as colunas e valores da TCC TIPOVIOL (arquivo TIPOVIOL.cnv)
    def get_TIPOVIOL_treated(self):
        # Conversão da TCC TIPOVIOL para um objeto pandas DataFrame
        file_name = 'TIPOVIOL'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'CIRCUNSTANCIA'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC FONTINFO (arquivo FONTINFO.cnv)
    def get_FONTINFO_treated(self):
        # Conversão da TCC FONTINFO para um objeto pandas DataFrame
        file_name = 'FONTINFO'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'ORIGEM'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC ATESTANT (arquivo ATESTANT.cnv)
    def get_ATESTANT_treated(self):
        # Conversão da TCC ATESTANT para um objeto pandas DataFrame
        file_name = 'ATESTANT'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'ATESTADOR'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC FONTEINV (arquivo FONTEINV.cnv)
    def get_FONTEINV_treated(self):
        # Conversão da TCC FONTEINV para um objeto pandas DataFrame
        file_name = 'FONTEINV'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'ORIGEM'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da TCC ESCAGR1 (arquivo ESCAGR1.cnv)
    def get_ESCAGR1_treated(self):
        # Conversão da TCC ESCAGR1 para um objeto pandas DataFrame
        file_name = 'ESCAGR1'
        df = download_table_cnv(file_name)
        # Renomeia a coluna "SIGNIFICACAO"
        df.rename(index=str, columns={'SIGNIFICACAO': 'ESCOLARIDADE'}, inplace=True)
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela TPOBITOCOR ("constando" apenas...
    # do Dicionário de Dados do SIM)
    def get_TPOBITOCOR_treated(self):
        # Conversão da Tabela TPOBITOCOR (em formato "xlsx") para um objeto pandas DataFrame
        df = pd.read_excel(self.path + 'TPOBITOCOR' + '.xlsx')
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'EPOCA_MORTE'}, inplace=True)
        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df



if __name__ == '__main__':

    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)

    # Ajustar o "the_path_..." para a localização dos arquivos "xlsx"
    the_path_Tabelas = os.getcwd()[:-len('\\insertion\\data_wrangling')] + 'files\\SIM\\'



###########################################################################################################################
# SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN #
###########################################################################################################################

import os
from datetime import datetime

import numpy as np
import pandas as pd

from transform.extract.download_SINAN import download_SINANXXaa, download_table_dbf, download_table_cnv


"""
Módulo de limpeza/tratamento de dados do SINAN.

"""

# Função para converter um "value" num certo "type" de objeto ou caso não seja possível utiliza o valor "default"
def tryconvert(value, default, type):
    try:
        return type(value)
    except (ValueError, TypeError):
        return default


# Classe de dados principais do SINAN
class DataSinanMain:

    # Construtor
    def __init__(self, base, state, year):
        self.base = base
        self.state = state
        self.year = year


    # Método para ler como um objeto pandas DataFrame um arquivo principal de dados do SINAN e adequar e formatar suas...
    # colunas e valores
    def get_SINANXXaamm_treated(self):
        # Lê o arquivo "dbc" ou "parquet", se já tiver sido baixado, como um objeto pandas DataFrame
        dataframe = download_SINANXXaa(self.base, self.state, self.year)
        print(f'O número de linhas do arquivo {self.base}{self.state}{self.year} é {dataframe.shape[0]}.')

        ###################################################################################################################
        # SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG #
        ###################################################################################################################
        if self.base == 'DENG':
            # Colunas definidas como necessárias no objeto pandas DataFrame que incrementará a tabela dengbr da base de dados
            lista_columns = np.array(['NU_NOTIFIC', 'TP_NOT', 'ID_AGRAVO', 'ID_MUNICIP', 'RES_CHIKS1', 'RES_CHIKS2',
                                      'RESUL_PRNT', 'RESUL_SORO', 'RESUL_NS1', 'SOROTIPO', 'HOSPITALIZ', 'CLASSI_FIN',
                                      'EVOLUCAO', 'GRAV_PULSO',  'GRAV_CONV', 'GRAV_ENCH',  'GRAV_INSUF', 'GRAV_TAQUI',
                                      'GRAV_EXTRE', 'GRAV_HIPOT', 'GRAV_HEMAT', 'GRAV_MELEN', 'GRAV_METRO', 'GRAV_SANG',
                                      'GRAV_AST', 'GRAV_MIOC', 'GRAV_CONSC', 'GRAV_ORGAO'])

            # Criação de um objeto pandas DataFrame vazio com as colunas especificadas acima
            df = pd.DataFrame(columns=lista_columns)

            # Colocação dos dados da variável "dataframe" na variável "df" nas colunas de mesmo nome preenchendo...
            # automaticamente com o float NaN as colunas da variável "df" não presentes na variável dataframe
            for col in df.columns.values:
                for coluna in dataframe.columns.values:
                    if coluna == col:
                        df[col] = dataframe[coluna].tolist()
                        break

            # Coloca na variável "dif_set" o objeto array dos nomes das colunas da variável "df" que não estão...
            # presentes na variável "dataframe"
            dif_set = np.setdiff1d(df.columns.values, dataframe.columns.values)

            # Substitui o float NaN pela string vazia as colunas da variável "df" não presentes na variável "dataframe"
            for col in dif_set:
                df[col].replace(np.nan, '', inplace=True)

            # Exclui o último dígito numérico das colunas identificadas, o qual corresponde ao dígito de...
            # controle do código do Município
            # Foi detectado que para alguns Municípios o cálculo do dígito de controle não é válido
            # Esse dígito de controle esteve presente nos arquivos DOXXxxxx até o ano de 2005 (a confirmar!)
            if len(df.loc[0, 'ID_MUNICIP']) == 7:
                df['ID_MUNICIP'].replace(regex='.$',value='', inplace=True)

            # Atualiza/corrige os labels das colunas especificadas
            df['ID_MUNICIP'] = df['ID_MUNICIP'].apply(lambda x: x if len(x) == 6 else '')
            df['ID_MUNICIP'].replace(['000000', '150475', '207540', '207695', '241005',
                                      '282580', '279982', '282586', '292586', '315205',
                                      '321213', '355038', '405028', '421265', '422000',
                                      '431454', '500627', '596382', '613167', '613592',
                                      '990010', '990014', '999999'], '', inplace=True)
            df['ID_MUNICIP'].replace([str(i) for i in range(334501, 334531)], '330455', inplace=True)
            df['ID_MUNICIP'].replace([str(i) for i in range(358001, 358059)], '355030', inplace=True)
            df['ID_MUNICIP'].replace(['530000', '530100', '530200', '530300', '530400',
                                      '530500', '530600', '530700', '530800', '530900',
                                      '531000', '531200', '531300', '531400', '531500',
                                      '531600', '531700', '531800'] + \
                                      [str(i) for i in range(539900, 540000)], '530010', inplace=True)

            df['CLASSI_FIN'].replace(['0', '6', '7', '9'], '', inplace=True)
            df['EVOLUCAO'].replace(['0', '9'], '', inplace=True)

            # Substitui uma string vazia pela string "NA" nas colunas de foreign keys
            for col in ['ID_MUNICIP', 'CLASSI_FIN']:
                df[col].replace('', 'NA', inplace=True)

            # Substitui uma string vazia por None nas colunas de atributos especificadas
            for col in ['TP_NOT', 'ID_AGRAVO', 'RES_CHIKS1', 'RES_CHIKS2', 'RESUL_PRNT',
                        'RESUL_SORO', 'RESUL_NS1', 'SOROTIPO', 'HOSPITALIZ', 'EVOLUCAO']:
                df[col].replace('', None, inplace=True)

            # Converte do tipo object para int ou para None as colunas de atributos de valores binários (0 ou 1)
            for col in np.array(['GRAV_PULSO', 'GRAV_CONV', 'GRAV_ENCH', 'GRAV_INSUF',
                                 'GRAV_TAQUI', 'GRAV_EXTRE', 'GRAV_HIPOT', 'GRAV_HEMAT',
                                 'GRAV_MELEN', 'GRAV_METRO', 'GRAV_SANG', 'GRAV_AST',
                                 'GRAV_MIOC', 'GRAV_CONSC', 'GRAV_ORGAO']):
                df[col] = df[col].apply(lambda x: tryconvert(x, None, int))

            # Renomeia colunas
            df.rename(index=str, columns={'ID_MUNICIP': 'MUNICIP_ID', 'CLASSI_FIN': 'CLASSIFIN_ID'}, inplace=True)

            print(f'Tratou o arquivo DENG{self.state}{self.year} (shape final: {df.shape[0]} x {df.shape[1]}).')

        return df


# Classe de dados auxiliares do SINAN
class DataSinanAuxiliary:

    # Construtor
    def __init__(self, path):
        self.path = path


    ###########################################################################################################################
    # SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN SINAN #
    ###########################################################################################################################




    ###################################################################################################################
    # SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG SINAN_DENG #
    ###################################################################################################################
    # Função para adequar e formatar as colunas e valores da Tabela TABUF (arquivo TABUF.dbf)
    def get_TABUF_treated(self):
        # Conversão da Tabela TABUF para um objeto pandas DataFrame
        file_name = 'TABUF'
        df = download_table_dbf(file_name)
        # Renomeia colunas especificadas
        df.rename(index=str, columns={'CODIGO': 'ID', 'DESCRICAO': 'ESTADO'}, inplace=True)
        # Reordena as colunas
        df = df[['ID', 'ESTADO', 'SIGLA_UF']]
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE', '?']
        return df


    # Função para adequar e formatar as colunas e valores da Tabela RSAUDE (do IBGE)
    def get_RSAUDE_treated(self):
        # Conversão da Tabela RSAUDE (em formato "xlsx") para um objeto pandas DataFrame
        df = pd.read_excel(self.path + 'RSAUDE' + '.xlsx')
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'REGIAO'}, inplace=True)
        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df


    # Método para adequar e formatar as colunas e valores da Tabela CADMUN (arquivo CADMUN.dbf)
    def get_CADMUN_treated(self):
        # Conversão da Tabela CADMUN para um objeto pandas DataFrame
        file_name = 'CADMUN'
        df1 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df1.rename(index=str, columns={'MUNCOD': 'ID', 'UFCOD': 'UFCOD_ID'}, inplace=True)
        # Drop a linha inteira em que a coluna "ID" tem o valor especificado por não representar nenhum município
        df1 = df1.drop(df1[df1['ID']=='000000'].index)
        # Remove colunas indesejáveis do objeto pandas DataFrame
        df1 = df1.drop(['MUNSINON', 'MUNSINONDV', 'MESOCOD', 'MICROCOD', 'MSAUDCOD',
                        'RSAUDCOD', 'CSAUDCOD', 'RMETRCOD', 'AGLCOD'], axis=1)
        # Substitui uma string vazia pela string "?" nas colunas especificadas
        for col in ['SITUACAO', 'MUNSINP', 'MUNSIAFI', 'MUNNOME', 'MUNNOMEX', 'OBSERV',
                    'AMAZONIA', 'FRONTEIRA', 'CAPITAL', 'ANOINST', 'ANOEXT', 'SUCESSOR']:
            df1[col].replace('', '?', inplace=True)
        # Substitui uma string vazia pela string "NA" nas colunas especificadas
        df1['UFCOD_ID'].replace('', 'NA', inplace=True)
        # Substitui uma string vazia pelo float "NaN" nas colunas especificadas
        for col in ['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']:
            df1[col].replace('', np.nan, inplace=True)
        # Converte do tipo object para float as colunas especificadas
        df1[['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']] = \
        df1[['LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA']].astype('float')
        # Coloca todas as string das colunas especificadas como UPPER CASE
        df1['MUNNOME'] = df1['MUNNOME'].apply(lambda x: x.upper())
        df1['MUNNOMEX'] = df1['MUNNOMEX'].apply(lambda x: x.upper())
        # Insere uma linha referente ao Município de Nazária/PI não constante originalmente do arquivo
        df1.loc[df1.shape[0]] = ['220672', '2206720', 'ATIVO', '?', '?', 'NAZÁRIA', 'NAZARIA', '?',
                                 'N', 'N', 'N', '22', '?', '?', '?', np.nan, np.nan, np.nan, 363.589]
        # Ordena as linhas de "df1" por ordem crescente dos valores da coluna ID
        df1.sort_values(by=['ID'], inplace=True)
        # Reset o index devido ao sorting prévio e à exclusão e inclusão das linhas referidas acima
        df1.reset_index(drop=True, inplace=True)
        # Conversão da Tabela rl_municip_regsaud para um objeto pandas DataFrame
        file_name = 'rl_municip_regsaud'
        df2 = download_table_dbf(file_name)
        # Renomeia as colunas especificadas
        df2.rename(index=str, columns={'CO_MUNICIP': 'ID', 'CO_REGSAUD': 'RSAUDE_ID'}, inplace=True)
        # Faz o merge de "df1" e "df2" pela coluna ID tendo por base "df1"
        df = pd.merge(df1, df2, how='left', left_on='ID', right_on='ID')
        # Converte o float NaN para a string "NA"
        df['RSAUDE_ID'].replace(np.nan, 'NA', inplace=True)
        # Reordena as colunas priorizando as "mais" relevantes
        df = df[['ID', 'MUNNOME', 'MUNNOMEX', 'MUNCODDV', 'OBSERV', 'SITUACAO', 'MUNSINP',
                 'MUNSIAFI', 'UFCOD_ID', 'AMAZONIA', 'FRONTEIRA', 'CAPITAL', 'RSAUDE_ID',
                 'LATITUDE', 'LONGITUDE', 'ALTITUDE', 'AREA', 'ANOINST', 'ANOEXT', 'SUCESSOR']]
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE', '?', '?', '?', '?', '?', '?', 'NA', '?',
                               '?', '?', 'NA', np.nan, np.nan, np.nan, np.nan, '?', '?', '?']
        return df


    # Função para adequar e formatar as colunas e valores da TCC Classdeng (arquivo Classdeng.cnv)
    def get_Classdeng_treated(self):
        # Conversão da TCC Classdeng para um objeto pandas DataFrame
        file_name = 'Classdeng'
        df = download_table_cnv(file_name)
        # Renomeia a coluna SIGNIFICACAO
        df.rename(index=str, columns={'SIGNIFICACAO': 'CLASSIFICACAO'}, inplace=True)

        # Converte para string a coluna especificada
        df['ID'] = df['ID'].astype('str')
        # Inserção de umas linhas no objeto pandas DataFrame
        df.loc[df.shape[0]] = ['10', 'DENGUE']
        df.loc[df.shape[0]] = ['11', 'DENGUE COM SINAIS DE ALARME']
        df.loc[df.shape[0]] = ['12', 'DENGUE GRAVE']
        df.loc[df.shape[0]] = ['13', 'CHIKUNGUNYA']
        # Inserção da primary key "NA" na tabela de que trata esta função para retratar "missing value"
        df.loc[df.shape[0]] = ['NA', 'NOT AVAILABLE']
        return df



if __name__ == '__main__':

    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)
